/****************************************/
/*  FileData.c headers         	        */
/*					*/
/*  Description : Datas attached t each */
/*		  file. 		*/
/*					*/
/*  Written by : Daniel Lacroix	        */
/*  GFile is under the GPL2 licence     */
/****************************************/

#include "GFile.h"

/****************************/
/* free a FileData          */
/* the file type is not free*/
void file_data_free(FileData *pFileData)
{
  pFileData->ref_count--;
  if(pFileData->ref_count == 0)
  {
    if(pFileData->path != NULL)         g_free(pFileData->path); 
    if(pFileData->name != NULL)         g_free(pFileData->name);
    if(pFileData->link_to_file != NULL) g_free(pFileData->link_to_file);
    if(pFileData->link_stat != NULL)    g_free(pFileData->link_stat);
    g_free(pFileData);
  }
}
/****************************/

/*******************************/
/* Return a copy of pFileData  */
FileData *file_data_copy(FileData *pFileData)
{
  pFileData->ref_count++;
  return pFileData;
}
/*******************************/

/**********************************************/
/* Return TRUE if p1FileData and p2FileData   */
/* indicate the same file. Else FALSE.        */
gboolean file_data_is_equal(FileData *p1FileData,FileData *p2FileData)
{
  if(strcmp(p1FileData->path,p2FileData->path)) return FALSE;
  if(strcmp(p1FileData->name,p2FileData->name)) return FALSE;
  return TRUE;
}
/**********************************************/

/********************************************************************/
/* create a new file data filled with the info of file "file_name"  */
/* in directory "file_path" (file_path is ALWAYS ended by a /).     */
/* The corresponding file_type is attached to the file if pFileType */
/* is not NULL (pFileType is a list of all the define file_type)    */
/* pOwner is the list of all the known user on this computer, if    */
/* NULL then we don't check the FILE_DATA_IS_ALLOW_TO_EXEC          */
/* The FileData returned can be NULL and GFileError can be set to   */
/* FILE_DATA_LSTAT_ERROR.                                           */
FileData *file_data_new_from_file(gchar *file_path,gchar *file_name,
  FileTypeList *pFileTypeList,Owner *pOwner)
{
  FileData *vFileData;
  gint  nb_car_in_link;
  guint link_buffer_size = 50;

  gchar *currentfich;
  
  vFileData = g_malloc(sizeof(FileData));

  currentfich = g_strconcat(file_path,file_name,NULL);

  if(lstat(currentfich,&(vFileData->stat)))
  {
    g_print("ATTENTION : stat error on file %s.\n",currentfich);
    gfile_errno = FILE_DATA_LSTAT_ERROR;
    return(NULL);
  }

  vFileData->path = g_strdup(file_path);
  vFileData->name = g_strdup(file_name);
  vFileData->link_stat    = NULL;
  vFileData->type         = NULL;
  vFileData->link_to_file = NULL;
  vFileData->extra_info   = 0;
  vFileData->ref_count    = 1;

  /* try to know if it is a link or not */
  if( S_ISLNK(vFileData->stat.st_mode) )
  {
    vFileData->link_stat = g_malloc(sizeof(struct stat));
    if(stat(currentfich,vFileData->link_stat))
    {
      vFileData->extra_info |= FILE_DATA_IS_BROKEN_SYMB;
      g_free(vFileData->link_stat);
      vFileData->link_stat = NULL;
    }
    
    vFileData->link_to_file = g_malloc(link_buffer_size);
    /* read the link name (try read while we have not enought to
      store the name string) */
    while ((nb_car_in_link = readlink(currentfich, vFileData->link_to_file, 
            link_buffer_size-1)) == link_buffer_size-1)
    { link_buffer_size += 30;
      g_free(vFileData->link_to_file);
      vFileData->link_to_file = g_malloc(link_buffer_size);
    }
      
    if(nb_car_in_link == -1) {
      g_free(vFileData->link_to_file);
      vFileData->link_to_file = g_strdup("can't follow link");
    } else
      /* ended the link name by a 0 */
      vFileData->link_to_file[nb_car_in_link] = 0;
  }

  /* we check if it is executable for somebody */
  if((vFileData->stat.st_mode & S_IXUSR)||(vFileData->stat.st_mode & S_IXGRP)||
     (vFileData->stat.st_mode & S_IXOTH))
  {
    vFileData->extra_info |= FILE_DATA_IS_EXECUTABLE;
    /* check if the gfile user can exec it */
    /* and if we want to check it */
    if(pOwner != NULL)
    {
      if((vFileData->stat.st_mode & S_IXOTH)||
	 ((vFileData->stat.st_mode & S_IXGRP) &&
	  (owner_current_user(pOwner)->grp->nb == vFileData->stat.st_gid) )||
	   ((vFileData->stat.st_mode & S_IXUSR) &&
	  (owner_current_user(pOwner)->nb ==  vFileData->stat.st_uid) ))
        vFileData->extra_info |= FILE_DATA_IS_ALLOW_TO_EXEC;
    }
  }

  g_free(currentfich);

  /* if we want to attach the corresponding FileType */
  if(pFileTypeList != NULL)
  {
    if(file_data_find_type(pFileTypeList,vFileData) == NULL)
    {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  }
  return(vFileData);
}
/********************************************************************/

/************************************************/
/* Find the filetype corresponding to pFileData */
/* and attach it to pFileData                   */
/* Return : The corresponding FileType.         */
FileType *file_data_find_type(FileTypeList *pFileTypeList, FileData *pFileData)
{ gboolean is_link;

  if(S_ISLNK(pFileData->stat.st_mode)) is_link=TRUE; else is_link=FALSE;

  /* if it is a broken symblink */
  if(S_ISLNK(pFileData->stat.st_mode) && 
     (pFileData->extra_info & FILE_DATA_IS_BROKEN_SYMB))
  {
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      BROKEN_LINK,FALSE);
    if(pFileData->type) return(pFileData->type);
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      BROKEN_LINK,TRUE);
    if(pFileData->type) return(pFileData->type);
    goto default_type;
  }

  /* a directory */   
  if((is_link  && S_ISDIR(pFileData->link_stat->st_mode)) ||
     (!is_link && S_ISDIR(pFileData->stat.st_mode)))
  {
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      NORMAL_DIRECTORY,is_link);
    if(pFileData->type == NULL)
    {g_print("NORMAL_DIRECTORY FileType not found.\n"); exit(FILE_TYPE_ERROR);}
    return(pFileData->type);
  }

  /* a character device */
  if((is_link  && S_ISCHR(pFileData->link_stat->st_mode)) ||
     (!is_link && S_ISCHR(pFileData->stat.st_mode)))
  {
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      CHARACTER_DEVICE,is_link);
    if(pFileData->type) return(pFileData->type);
  }

  /* a block device */
  if((is_link  && S_ISBLK(pFileData->link_stat->st_mode)) ||
     (!is_link && S_ISBLK(pFileData->stat.st_mode)))
  {
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      BLOCK_DEVICE,is_link);
    if(pFileData->type) return(pFileData->type);
  }

  /* a fifo */
  if((is_link  && S_ISFIFO(pFileData->link_stat->st_mode)) ||
     (!is_link && S_ISFIFO(pFileData->stat.st_mode)))
  {
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      FIFO,is_link);
    if(pFileData->type) return(pFileData->type);
  }

  /* a socket */
  if((is_link  && S_ISSOCK(pFileData->link_stat->st_mode)) ||
     (!is_link && S_ISSOCK(pFileData->stat.st_mode)))
  {
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      SOCKET,is_link);
    if(pFileData->type) return(pFileData->type);
  }

  /* if this is a link on an executable */
  if(is_link && S_ISREG(pFileData->link_stat->st_mode) &&
     (pFileData->link_stat->st_mode && (S_IXUSR|S_IXGRP|S_IXOTH) ))
  {
    pFileData->type = file_type_list_find_file_type(pFileTypeList,
      EXECUTABLE,is_link);
    if(pFileData->type) return(pFileData->type);
  }

  /* a regular file */
  if(S_ISREG(pFileData->stat.st_mode))
  {
    pFileData->type = file_type_list_find_user_defined_file_type(
     pFileTypeList,pFileData->name,is_link);
    if(pFileData->type != NULL) /* we found a corresponding type */
      return(pFileData->type);

    /* perhaps an executable ?*/
    if(pFileData->extra_info & FILE_DATA_IS_EXECUTABLE)
    {
      pFileData->type = file_type_list_find_file_type(pFileTypeList,
        EXECUTABLE,is_link);
      if(pFileData->type) return(pFileData->type);
    }
  }

default_type:

  /* not define */
  pFileData->type = file_type_list_find_file_type(pFileTypeList,
    DEFAULT_TYPE,is_link);
  if(pFileData->type == NULL)
  {g_print("DEFAULT_TYPE FileType not found.\n"); exit(FILE_TYPE_ERROR);}
  return(pFileData->type);
}
/************************************************/

/********************************************************************/
/* Scan the content of a directory. Create a FileData for each      */
/* entries. For eached FileData, user_func is called with the       */
/* FileData and it own data. If user_func return FALSE the FileData */
/* is suppress.                                                     */
/* The corresponding file_type is attached to the file if pFileType */
/* is not NULL (pFileType is a list of all the define file_type)    */
/* pOwner is the list of all the known user on this computer, if    */
/* NULL then we don't check the FILE_DATA_IS_ALLOW_TO_EXEC.         */
/* If file_data_open_dir return FALSE, gfile_errno can be set to :  */
/*  FILE_DATA_DIR_EACCES  Permission denied.                                   */
/*  FILE_DATA_DIR_EMFILE  Too many file descriptors in use by process.         */
/*  FILE_DATA_DIR_ENFILE  Too many files are currently open in the system.     */
/*  FILE_DATA_DIR_ENOENT  Directory does not exist, or name is an empty string.*/
/*  FILE_DATA_DIR_ENOMEM  Insufficient memory to complete the operation.       */ 
/*  FILE_DATA_DIR_ENOTDIR name is not a directory.                             */
/*  FILE_DATA_DIR_UNKNOWN unknown error.                                       */
gboolean file_data_open_dir(gchar *dir_name,
  FileTypeList *pFileTypeList,Owner *pOwner,
  gboolean (* user_func)(FileData *pFileData,gpointer data),gpointer data)
{
  DIR *directory;
  gchar *string;
  struct dirent *file_dirent;
  FileData  *vFileData;
  GtkWidget *error_dialog;
  gchar     *error_msg;

  directory = opendir(dir_name);
      
  if(directory == NULL)
  {
    switch(errno)
    {
      case EACCES  : gfile_errno = FILE_DATA_DIR_EACCES;  break;
      case EMFILE  : gfile_errno = FILE_DATA_DIR_EMFILE;  break;
      case ENFILE  : gfile_errno = FILE_DATA_DIR_ENFILE;  break;
      case ENOENT  : gfile_errno = FILE_DATA_DIR_ENOENT;  break;
      case ENOMEM  : gfile_errno = FILE_DATA_DIR_ENOMEM;  break;
      case ENOTDIR : gfile_errno = FILE_DATA_DIR_ENOTDIR; break;      
      default      : gfile_errno = FILE_DATA_DIR_UNKNOWN; break;
    }
    return(FALSE);
  }

  /* if we are not scanning the root, add a '/' to make the path */
  if(strcmp(dir_name,"/"))
    string = g_strconcat(dir_name,"/",NULL);
  else
    string = g_strdup(dir_name);
    
  while((file_dirent = readdir(directory)) != NULL)
  {
    vFileData = file_data_new_from_file(string,file_dirent->d_name,
      pFileTypeList,pOwner);
    if(vFileData == NULL)
    {
      error_msg = g_strconcat("error while stating file \"",
        file_dirent->d_name,"\".",NULL);
      error_dialog = gnome_message_box_new(error_msg,GNOME_MESSAGE_BOX_ERROR,
        GNOME_STOCK_BUTTON_OK,NULL);
      g_free(error_msg);
      gtk_grab_add(error_dialog);
      gtk_signal_connect(GTK_OBJECT(error_dialog),"destroy",
        GTK_SIGNAL_FUNC(gtk_grab_remove),NULL);
      gtk_widget_show(error_dialog);
    } else {
      if(user_func(vFileData,data) == FALSE) file_data_free(vFileData);
    }
  }
  g_free(string);

  closedir(directory);
  return(TRUE);
}
/********************************************************************/
