/****************************************/
/*  FileType.c body                     */
/*					*/
/*  Description : FileType take care of */
/*		  file type reconization*/
/*					*/
/*  Written by : Daniel Lacroix	        */
/*  GFile is under the GPL2 licence     */
/****************************************/

#include "GFile.h"

/* usefull for gdk_pixmap_create_from_xpm_d in load_special_type */
static GdkWindow *parent_window_app = NULL;
/* the icon path when not full path define */
gchar *icon_path = "/usr/share/pixmaps/";
static gboolean icon_path_never_set = TRUE;

/* string corresponding to mode enum */
gchar *mode_str[]={
     "NORMAL_DIRECTORY",
     "NORMAL_OPEN_DIRECTORY",
     "CURRENT_DIRECTORY",
     "CURRENT_OPEN_DIRECTORY",
     "PAST_DIRECTORY",
     "PAST_OPEN_DIRECTORY",
     "DEFAULT_TYPE",
/* MUST BE PRESENT TYPE ARE BETWEEN 1 AND LAST_MUST_BE_TYPE */
     "error not defined",
/* facultative type */
     "BROKEN_LINK",
     "EXECUTABLE",
     "SOCKET",
     "FIFO",
     "BLOCK_DEVICE",
     "CHARACTER_DEVICE",
     "USER_DEFINED"};

#include "FileTypeInternalIcons.h"

/****************************************************************/
/* set a new icon_path for icon not full path define            */
/* exemple of path : /usr/share/pixmap/ (always ended by a '/') */
void file_type_set_icon_path(gchar *new_path)
{
  if(icon_path_never_set)
    icon_path_never_set = FALSE;
  else
    g_free(icon_path);
  icon_path = g_strdup(new_path);
}
/****************************************************************/

/***************************************************************/
/* Create a new file type.                                     */
/* Return : NULL if impossible and gfile_errno can be set to : */
/* FILE_TYPE_PARAM_ERROR                                       */
/* FILE_TYPE_UNKOW_MUST_BE_FILE_TYPE                           */
/* FILE_TYPE_LOADING_ICON_ERROR                                */

/* POSSIBLE IMPROVEMENT : define the icon_path as /icons/:/usr/icons/:...*/
FileType *file_type_new(guint8 mode, gchar *type_str, gchar *program_str,
  gchar *icon_name, gboolean is_link, GdkPixmap *pixmap, GdkBitmap *mask)
{ FileType *vFileType;
  gchar **vXpmPixmap;
  gchar *full_icon_name;
  GdkImlibImage *pix;  

  if(((pixmap == NULL) && (mask != NULL)) || ((pixmap != NULL) && (mask == NULL)))
  {
    gfile_errno = FILE_TYPE_PARAM_ERROR;
    return(NULL);
  }
 
  /* if this is not a user defined type, type_str and program_str MUST be NULL */
  if(mode == USER_DEFINED)
  {
    if((type_str == NULL) || (icon_name == NULL))
    {     
      gfile_errno = FILE_TYPE_PARAM_ERROR;
      return(NULL);
    }
  } else if((type_str != NULL) || (program_str != NULL))
  {
    gfile_errno = FILE_TYPE_PARAM_ERROR;
    return(NULL);
  }
  /* alloc mem for the new file type */
  vFileType = g_malloc(sizeof(FileType));
  
  if(type_str != NULL)
    vFileType->type_str = g_strdup(type_str);
  else
    vFileType->type_str = NULL;
  
  if(program_str != NULL)
    vFileType->program_str = g_strdup(program_str);
  else  
    vFileType->program_str = NULL;

  if(icon_name != NULL)
    vFileType->icon_name = g_strdup(icon_name);
  else
    vFileType->icon_name = NULL;

  vFileType->is_link = is_link;
  
  vFileType->mode = mode;

  if(pixmap != NULL)
  { /* icon pixmap is given in parameter */
    vFileType->pixmap = gdk_pixmap_ref(pixmap);
    vFileType->mask   = gdk_bitmap_ref(mask);    
  } else if(icon_name != NULL)
  { /* icon pixmap must be load from the hard drive */
    if(icon_name[0] != '/')
      full_icon_name = g_strconcat(icon_path,icon_name,NULL);
    else
      full_icon_name = icon_name;  
  
    pix = gdk_imlib_load_image(full_icon_name);
    if(pix == NULL)
    {
      g_free(vFileType->icon_name); g_free(vFileType->program_str);
      g_free(vFileType->type_str); g_free(vFileType);
      gfile_errno = FILE_TYPE_LOADING_ICON_ERROR;
      return(NULL);
    }
/*    gdk_imlib_render (pix, pix->rgb_width, pix->rgb_height);*/
    gdk_imlib_render (pix, 16, 16);
    vFileType->pixmap = gdk_imlib_copy_image(pix);
    vFileType->mask   = gdk_imlib_copy_mask(pix);
    gdk_imlib_destroy_image(pix);
    
    if(icon_name[0] != '/') g_free(full_icon_name);
  } else {
    /* internal icons will be used if present */
    if(mode > LAST_MUST_BE_TYPE)
    { /* not a "must be" type, no internal icons */
      g_free(vFileType->icon_name); g_free(vFileType->program_str);
      g_free(vFileType->type_str); g_free(vFileType);
      gfile_errno = FILE_TYPE_PARAM_ERROR;
      return(NULL);
    }
    
    switch(mode)
    {
      case NORMAL_DIRECTORY       : vXpmPixmap = NORMAL_DIRECTORY_XPM;       break;
      case NORMAL_OPEN_DIRECTORY  : vXpmPixmap = NORMAL_OPEN_DIRECTORY_XPM;  break;
      case CURRENT_DIRECTORY      : vXpmPixmap = CURRENT_DIRECTORY_XPM;      break;
      case CURRENT_OPEN_DIRECTORY : vXpmPixmap = CURRENT_OPEN_DIRECTORY_XPM; break;
      case PAST_DIRECTORY         : vXpmPixmap = PAST_DIRECTORY_XPM;         break;
      case PAST_OPEN_DIRECTORY    : vXpmPixmap = PAST_OPEN_DIRECTORY_XPM;    break;
      case DEFAULT_TYPE           : vXpmPixmap = DEFAULT_TYPE_XPM;           break;
      default : 
        g_free(vFileType->icon_name); g_free(vFileType->program_str);
        g_free(vFileType->type_str); g_free(vFileType); 
	gfile_errno = FILE_TYPE_UNKOW_MUST_BE_FILE_TYPE;
	return(NULL); break;
    }
      
    vFileType->pixmap = gdk_pixmap_create_from_xpm_d(
      parent_window_app,&vFileType->mask,NULL,vXpmPixmap);			 
  }
  
  return(vFileType);
}
/***************************************************************/

/*********************/
/* Free a file type. */
void file_type_free(FileType *pFileType)
{
  if(pFileType == NULL) return;

  if(pFileType->type_str != NULL)    g_free(pFileType->type_str);
  if(pFileType->program_str != NULL) g_free(pFileType->program_str);
  if(pFileType->icon_name != NULL)   g_free(pFileType->icon_name);

  /* free the pixmap attached to the type */
  gdk_pixmap_unref(pFileType->pixmap);
  /* free the mask attached to the type */
  gdk_bitmap_unref(pFileType->mask);
  /* free the type structure */
  g_free(pFileType);
}
/*********************/

/**********************************************************/
/* set the parent application to avoid warning when using */
/* gdk_pixmap_create_from_xpm_d                           */
void file_type_set_parent(GtkWidget *parent)
{ parent_window_app = parent->window; }
/**********************************************************/

/*********************************************************/
/* Return a copy of pFileType                            */
/* or NULL if impossible and gfile_errno can be set to : */
/* FILE_TYPE_PARAM_ERROR                                 */
/* FILE_TYPE_UNKOW_MUST_BE_FILE_TYPE                     */
/* FILE_TYPE_LOADING_ICON_ERROR                          */
FileType *file_type_copy(FileType *pFileType)
{
  if(pFileType == NULL)
  {
    gfile_errno = FILE_TYPE_PARAM_ERROR;
    return(NULL);
  }
  return(file_type_new(pFileType->mode,pFileType->type_str,
    pFileType->program_str,pFileType->icon_name,
    pFileType->is_link,pFileType->pixmap,pFileType->mask));
}
/*********************************************************/

/*******************************************************************/
/* Return the mode corresponding to mode string (NORMAL_DIRECTORY, */
/* NORMAL_OPEN_DIRECTORY..) or -1 if not found                     */
guint8 file_type_str_2_mode(gchar *mode_string)
{ guint8 mode;

  for (mode = NORMAL_DIRECTORY; mode <= USER_DEFINED; mode++)
  {
    if((mode != LAST_MUST_BE_TYPE) && (!strcmp(mode_str[mode],mode_string)))
      return(mode);
  }
  return(-1);
}
/*******************************************************************/
