/**********************************************/
/*  FileTypeList.c body                       */
/*					      */
/*  Description : FileTypeList take care of   */
/*		  file type reconization list */
/*					      */
/*  Written by : Daniel Lacroix	              */
/*  GFile is under the GPL2 licence           */
/**********************************************/

#include "GFile.h"

/*******************************************/
/* Create a new file type system.          */
/* Return : the new allocated FileTypeList */
FileTypeList *file_type_list_new(void)
{ FileTypeList *vFileTypeList;

  vFileTypeList = g_malloc(sizeof(FileTypeList));
  vFileTypeList->list =  g_slist_alloc();
  return(vFileTypeList);
}
/*******************************************/

/****************************/
/* Free a file type system. */
void file_type_list_free(FileTypeList *pFileTypeList)
{
  file_type_list_init(pFileTypeList);
  while(file_type_list_is_end(pFileTypeList) == FALSE)
  {
    file_type_free(file_type_list_current(pFileTypeList));
    file_type_list_next(pFileTypeList);
  }
  g_slist_free(pFileTypeList->list);
  g_free(pFileTypeList);
}
/****************************/

/***************************************/
/* insert after "before" in the list   */
/* if before = NULL, insert at the end */
void file_type_list_insert_file_type(FileTypeList *pFileTypeList,FileType *before,FileType *pfile_type)
{ GSList *last_list;
  GSList *new_list;

  new_list = g_malloc(sizeof(GSList));
  new_list->data = (gpointer)pfile_type;

  last_list = pFileTypeList->list;

  file_type_list_init(pFileTypeList);
  while((file_type_list_is_end(pFileTypeList) == FALSE) && 
        (file_type_list_current(pFileTypeList) != before))
  { 
    last_list = pFileTypeList->current;
    file_type_list_next(pFileTypeList);
  }

  if((before != NULL) && (file_type_list_current(pFileTypeList) == before))
  { 
    last_list = pFileTypeList->current;
    file_type_list_next(pFileTypeList);
  }

  new_list->next = pFileTypeList->current;
  last_list->next = new_list;
}
/***************************************/

/************************************************/
/* Remove a file type to the file type list     */
void file_type_list_remove_file_type(FileTypeList *pFileTypeList,
  FileType *remove_type)
{   
  GSList *last_type;
  
  last_type = pFileTypeList->list;
  
  if((pFileTypeList == NULL) || (remove_type == NULL)) return;
  
  /* free the element in the list*/
  file_type_list_init(pFileTypeList);
  while((file_type_list_is_end(pFileTypeList) == FALSE) &&
        (file_type_list_current(pFileTypeList) != remove_type))
  {
    last_type = pFileTypeList->current;
    file_type_list_next(pFileTypeList);
  }
  if(file_type_list_is_end(pFileTypeList) == TRUE)
    return; /* type not present */
  else {  
   last_type->next = pFileTypeList->current->next;
    g_free(pFileTypeList->current);
  }    
  
  file_type_free(remove_type);
}    
/************************************************/

/*****************************************************************/
/* swap two FileTypeList.                                        */
/* You CAN'T CALLED THIS IF YOU ARE USING file_type_list_next... */
void file_type_list_swap_file_type_list(FileTypeList *pFileTypeList1,
  FileTypeList *pFileTypeList2)
{ GSList *tmp;
 
  tmp = pFileTypeList1->list;
  pFileTypeList1->list = pFileTypeList2->list;
  pFileTypeList2->list = tmp;
}
/*****************************************************************/

/******************************************/
/* swap two file type in a file type list */
void file_type_list_swap_file_type(FileTypeList *pFileTypeList,FileType *type1,FileType *type2)
{ GSList *data1_list,*data2_list;

  /* find type1 in the list */  
  file_type_list_init(pFileTypeList);
  while((file_type_list_is_end(pFileTypeList) == FALSE) && 
        (file_type_list_current(pFileTypeList) != type1))
  { file_type_list_next(pFileTypeList); }
  /* type not found */
  if(file_type_list_is_end(pFileTypeList) == TRUE) return;
  data1_list = pFileTypeList->current; 
  /* find type2 in the list */    
  file_type_list_init(pFileTypeList);
  while((file_type_list_is_end(pFileTypeList) == FALSE) && 
        (file_type_list_current(pFileTypeList) != type2))
  { file_type_list_next(pFileTypeList); }
  /* type not found */
  if(file_type_list_is_end(pFileTypeList) == TRUE) return;
  data2_list = pFileTypeList->current;
  /* data both exist => swap */
  data2_list->data = (gpointer)type1; data1_list->data = (gpointer)type2;
}
/******************************************/

/*******************************************************/
/* add all the file types of psrc at the end of pdest. */
/* If pdest is empty, this make a copy of psrc.        */
void file_type_list_append_file_type_list(FileTypeList *pdest,FileTypeList *psrc)
{ GSList *list_node;

  /* security check */
  if((pdest == NULL) || (psrc == NULL)) return;

  file_type_list_init(psrc);
  while(file_type_list_is_end(psrc) == FALSE)
  {
    file_type_list_append_file_type(pdest,file_type_copy(file_type_list_current(psrc)));
    file_type_list_next(psrc);
  }
}
/*******************************************************/

/****************************************/
/* Return a copy of psrc file type list */
FileTypeList *file_type_list_copy(FileTypeList *psrc)
{ FileTypeList *vdest;

  vdest = file_type_list_new();
  file_type_list_append_file_type_list(vdest,psrc);
  return(vdest);
}
/****************************************/

/*********************************************/
/* Return the FileType corresponding to mode */
/* link or not or return NULL if not found   */
FileType *file_type_list_find_file_type(FileTypeList *pFileTypeList,guint8 mode,gboolean is_link)
{
  file_type_list_init(pFileTypeList);
  while((file_type_list_is_end(pFileTypeList) == FALSE) && 
        ((file_type_list_current(pFileTypeList)->is_link != is_link) ||
         (file_type_list_current(pFileTypeList)->mode != mode)))
  {
    file_type_list_next(pFileTypeList);
  }
  if(file_type_list_is_end(pFileTypeList) == FALSE)
    return(file_type_list_current(pFileTypeList));
  else
    return(NULL);
}
/*********************************************/

/*****************************************************/
/* Return the corresponding USER_DEFINER type for    */
/* filename link or not. Return NULL if none defined */
FileType *file_type_list_find_user_defined_file_type(FileTypeList *pFileTypeList,gchar *filename,gboolean is_link)
{
  file_type_list_init(pFileTypeList);
  while(file_type_list_is_end(pFileTypeList) == FALSE)
  {
    if((file_type_list_current(pFileTypeList)->mode == USER_DEFINED) &&
       (file_type_list_current(pFileTypeList)->is_link == is_link) &&
       is_of_multi_type(file_type_list_current(pFileTypeList)->type_str, filename))
       return(file_type_list_current(pFileTypeList));
    file_type_list_next(pFileTypeList);
  }
  return(NULL);
}
/*****************************************************/

/************************************************/
/* Save the file type list in the filetype file */
/* filename is the full path name of the file   */
/* Return : FALSE = impossible; TRUE = job done */
/* if FALSE, gfile_errno can be set to          */
/*   FILE_TYPE_LIST_CREAT_FILE_ERROR            */
/*   FILE_TYPE_LIST_WRITE_FILE_ERROR            */
gboolean file_type_list_save_to_file(FileTypeList *pFileTypeList,gchar *filename)
{
  gint file_handle;
  FileType *tmptype;
  gchar *string,*string2;

  file_handle = creat(filename,S_IREAD|S_IWRITE);
  if(file_handle == -1)
  { gfile_errno = FILE_TYPE_LIST_CREAT_FILE_ERROR; return(FALSE); }
  
  file_type_list_init(pFileTypeList);

  if(write(file_handle,"# syntaxe is : mode,type,pgm,icon,is_link.\n",
    strlen("# syntaxe is : mode,type,pgm,icon,is_link.\n")) == -1)
  { close(file_handle); gfile_errno = FILE_TYPE_LIST_WRITE_FILE_ERROR; return(FALSE); }
  while(file_type_list_is_end(pFileTypeList) == FALSE)
  {
    tmptype = file_type_list_current(pFileTypeList);
    /* if this is not an internal file_type we save it to the disk */
    if( tmptype->icon_name != NULL )
    {
      string = g_strconcat(mode_str[tmptype->mode],",",NULL);
      if(tmptype->type_str != NULL)
        string2 = g_strconcat(string,tmptype->type_str,",",NULL);
       else
        string2 = g_strconcat(string,",",NULL);
      g_free(string);
      string = string2;

      if(tmptype->program_str != NULL)
        string2 = g_strconcat(string,tmptype->program_str,",",NULL);
      else
        string2 = g_strconcat(string,",",NULL);      
      g_free(string);
      string = string2;

      string2 = g_strconcat(string,tmptype->icon_name,",",NULL);
      g_free(string);
      string = string2;

      if(tmptype->is_link == TRUE)
        string2 = g_strconcat(string,"TRUE.\n",NULL);
      else
        string2 = g_strconcat(string,"FALSE.\n",NULL);
      g_free(string);
      string = string2;

      if(write(file_handle,string,strlen(string)) == -1)
      { close(file_handle); gfile_errno = FILE_TYPE_LIST_WRITE_FILE_ERROR; return(FALSE); }      
      g_free(string);
    }
    /* next file type */
    file_type_list_next(pFileTypeList);
  }      
  
  close(file_handle);
  return(TRUE);
}
/************************************************/

/******************************************************/
/* Load the file type list from a file.               */
/* filename is the full path name of the file         */
/* Return : NULL = impossible gfile_errno can be set  */
/*          to FILE_TYPE_LIST_FILE_TOO_BIG            */
/*             FILE_TYPE_LIST_FILE_OPEN_ERROR         */
/*             FILE_TYPE_LIST_FILE_SYNTAXE_ERROR      */
/*            FILE_TYPE_LIST_FILE_INVALID_FILE_TYPE   */
/*            FILE_TYPE_LIST_FILE_ICON_NOT_FOUND      */
/*          else job done and the file type list fill */
FileTypeList *file_type_list_load_from_file(gchar *filename)
{
  FileTypeList *vFileTypeList;
  gint         file_handle;
  FileType     *vFileType;
  gchar        *buf;
  gint         buf_size;
  guint        buf_pos;
  guint        line_nb = 1;
  guint        line_begin = 0;
  guint        info_begin;
  guint        info_end;
  gboolean     info_in_fill = FALSE;
  gboolean     info_nothing_fill = TRUE;
  guint8       info_num = 0;
  gchar        tmp_char;

  guint8   mode;  
  gchar    *type_str    = NULL;  
  gchar    *program_str = NULL;
  gchar    *icon_name   = NULL; 
  gboolean is_link;
  GdkPixmap *pixmap    = NULL;
  GdkBitmap *mask      = NULL;
  
  vFileTypeList = file_type_list_new();

  /* if filename is NULL, then we just want to use internal type */

  if(filename != NULL) /* If the file exist, we use it */
  {
    /* try to open the file */
    file_handle = open(filename,O_RDONLY);
    if(file_handle == -1)
    {
      file_type_list_free(vFileTypeList);
      gfile_errno = FILE_TYPE_LIST_FILE_OPEN_ERROR;
      return(NULL);
    }
    /* alloc a bloc of memory of the file size */
    buf_size = lseek(file_handle, 0, SEEK_END) + 1;
    if(buf_size == 0)
    {
      file_type_list_free(vFileTypeList);
      close(file_handle);
      gfile_errno = FILE_TYPE_LIST_FILE_OPEN_ERROR;
      return(NULL);
    }
    /* place the file handle at the begin */
    if(lseek(file_handle, 0, SEEK_SET) == -1)
    {
      file_type_list_free(vFileTypeList);
      close(file_handle);
      gfile_errno = FILE_TYPE_LIST_FILE_OPEN_ERROR;
      return(NULL);
    }
    if(buf_size-1 > FILE_TYPE_LIST_MAX_FILE_SIZE)
    {
      g_print("FileTypeList : file \"%s\" is too big, it might be a mistake.\n");
      file_type_list_free(vFileTypeList);
      close(file_handle);
      gfile_errno = FILE_TYPE_LIST_FILE_TOO_BIG;
      return(NULL);
    }
    buf = g_malloc(buf_size);
    buf[buf_size-1] = 0;
    if(read(file_handle,buf,buf_size-1) != buf_size-1)
    {
      g_free(buf);
      file_type_list_free(vFileTypeList);
      close(file_handle);
      gfile_errno = FILE_TYPE_LIST_FILE_OPEN_ERROR;
      return(NULL);
    }
    close(file_handle);
    /* replace all line feed by 0 */
    for(buf_pos = 0; buf_pos < buf_size; buf_pos++)
      if(buf[buf_pos]=='\n') buf[buf_pos] = 0;
    /* add the file type define by the user in filetype file */

    buf_pos = 0;
    while( buf_pos < buf_size )
    {
      switch(buf[buf_pos])
      {
         case '#' : while(buf[++buf_pos] != 0);
	            break;
	 case 0   : line_nb++; buf_pos++;
	 	    line_begin = buf_pos;
	 	    /*g_print("line %u : \"%s\".\n",line_nb,&buf[line_begin]);*/
	            break;
	case '\t' :	    
	case ' '  : if(info_in_fill){info_end = buf_pos; info_in_fill = FALSE;}
		    buf_pos++;
	            break;
	case ','  : if(info_in_fill){info_end = buf_pos;  info_in_fill = FALSE;}
     	            buf_pos++;
		    switch(info_num)
		    {
		      case 0 : /* mode USER_DEFINED... */
		        if(info_nothing_fill)
			{
		          g_print("FileTypeList ERROR in file %s, line %u : \"%s\","\
		  	  " mode (USER_DEFINED...) NOT DEFINE.\n",
			  filename,line_nb,&buf[line_begin]);
			  file_type_list_free(vFileTypeList);
			  g_free(buf);
                          if(type_str)    g_free(type_str);
                          if(program_str) g_free(program_str);
                          if(icon_name)   g_free(icon_name);	  
			  gfile_errno = FILE_TYPE_LIST_FILE_SYNTAXE_ERROR;
			  return(NULL);
		        } else {
 		          buf[info_end] = 0;
		          mode = file_type_str_2_mode(&buf[info_begin]);
		          buf[info_end] = ',';
		 	  if(mode == 255)
			  {
		            g_print("FileTypeList ERROR in file %s, line %u : \"%s\","\
  			    " unknown mode (expected USER_DEFINED...).\n",
			    filename,line_nb,&buf[line_begin]);
			    file_type_list_free(vFileTypeList);
			    g_free(buf);
                            if(type_str)    g_free(type_str);
                            if(program_str) g_free(program_str);
                            if(icon_name)   g_free(icon_name);    
			    gfile_errno = FILE_TYPE_LIST_FILE_SYNTAXE_ERROR;
			    return(NULL);
			  } 
			}    
		        break;
		      case 1 : /* type string (for example : *.ps) */
		        if(info_nothing_fill) {
			  type_str = NULL;
		        } else {		      
		          buf[info_end] = 0;
		          type_str = g_strdup(&buf[info_begin]);
		  	  buf[info_end] = ',';
			}
		        break;		      
		      case 2 : /* program associated */
		        if(info_nothing_fill) {
			  program_str = NULL;
		        } else {
		          buf[info_end] = 0;
		          program_str = g_strdup(&buf[info_begin]);
		  	  buf[info_end] = ',';
			}
		        break;
		      case 3 : /* corresponding icon name */
		        if(info_nothing_fill)
			{
		          g_print("FileTypeList ERROR in file %s, line %u : \"%s\","\
		  	  " icon name NOT DEFINE.\n",filename,line_nb,&buf[line_begin]);
			  file_type_list_free(vFileTypeList);
			  g_free(buf);
                          if(type_str)    g_free(type_str);
                          if(program_str) g_free(program_str);
                          if(icon_name)   g_free(icon_name);	  
			  gfile_errno = FILE_TYPE_LIST_FILE_SYNTAXE_ERROR;
			  return(NULL);
		        } else {
 		          buf[info_end] = 0;
			  icon_name = g_strdup(&buf[info_begin]);
		          buf[info_end] = ',';
			}
		        break;
		      default :
   		        g_print("FileTypeList ERROR in file %s, line %u : \"%s\","\
		  	  " a type is ended by a '.'.\n",filename,line_nb,&buf[line_begin]);
			file_type_list_free(vFileTypeList);
			g_free(buf);
                        if(type_str)    g_free(type_str);
                        if(program_str) g_free(program_str);
                        if(icon_name)   g_free(icon_name);			
			gfile_errno = FILE_TYPE_LIST_FILE_SYNTAXE_ERROR;
			return(NULL);
                        break;
		    }
		    info_num++;
		    info_nothing_fill = TRUE;		    
		    break;
	case '.'  : if(info_num == 4)
	            { if(info_in_fill){info_end = buf_pos; info_in_fill = FALSE;}
		      buf_pos++;
  		      if(info_nothing_fill)
		      {
		        g_print("FileTypeList ERROR in file %s, line %u :\"%s\","\
			  " is a link NOT DEFINE.\n",filename,line_nb,&buf[line_begin]);
			file_type_list_free(vFileTypeList);
			g_free(buf);
                        if(type_str)    g_free(type_str);
                        if(program_str) g_free(program_str);
                        if(icon_name)   g_free(icon_name);			
			gfile_errno = FILE_TYPE_LIST_FILE_SYNTAXE_ERROR;
		        return(NULL);
		      } else {
		        if(!strncmp(&buf[info_begin],"FALSE",5)) {
			  is_link = FALSE;
			} else if(!strncmp(&buf[info_begin],"TRUE",4)) {
			  is_link = TRUE;
			} else {
		          g_print("FileTypeList ERROR in file %s, line %u : \"%s\","\
  			   " is a link MUST be TRUE or FALSE.\n",
			  filename,line_nb,&buf[line_begin]);
			  file_type_list_free(vFileTypeList);
 			  g_free(buf);
                          if(type_str)    g_free(type_str);
                          if(program_str) g_free(program_str);
                          if(icon_name)   g_free(icon_name);		  
			  gfile_errno = FILE_TYPE_LIST_FILE_SYNTAXE_ERROR;
			  return(NULL);			  
			}
		      }
                      info_num = 0;
		      info_nothing_fill = TRUE;
		      /* add it to the list */
		      file_type_list_find_icon(vFileTypeList, icon_name, 
		        &pixmap, &mask);
		      vFileType = file_type_new(mode,type_str,program_str,
		        icon_name,is_link,pixmap,mask);
                      if(type_str)    {g_free(type_str);    type_str    = NULL;}
                      if(program_str) {g_free(program_str); program_str = NULL;}
                      if(icon_name)   {g_free(icon_name);   icon_name   = NULL;}
		      if(vFileType == NULL)
		      {  
		        gdk_pixmap_unref(pixmap);
                        gdk_bitmap_unref(mask);
			file_type_list_free(vFileTypeList);
 			g_free(buf);
			if(gfile_errno == FILE_TYPE_PARAM_ERROR)
                          gfile_errno = FILE_TYPE_LIST_FILE_INVALID_FILE_TYPE;
                        if(gfile_errno == FILE_TYPE_LOADING_ICON_ERROR)
			  gfile_errno = FILE_TYPE_LIST_FILE_ICON_NOT_FOUND;
			return(NULL);
		      }
		      pixmap = NULL; mask = NULL;
                      file_type_list_append_file_type(vFileTypeList,vFileType);
		    } else goto normal_char;
		    break;

	 default  : 
	 normal_char :
	   if((info_in_fill == FALSE) && (info_nothing_fill == FALSE))
	   {
	     if(info_num == 2)
	     {
	       info_in_fill = TRUE;
	     } else {
               g_print("FileTypeList ERROR in file %s, line %u : \"%s\","\
                 " too much parameter.\n",filename,line_nb,&buf[line_begin]);
	       file_type_list_free(vFileTypeList);
	       g_free(buf);
               if(type_str)    g_free(type_str);
               if(program_str) g_free(program_str);
               if(icon_name)   g_free(icon_name);	     
	       gfile_errno = FILE_TYPE_LIST_FILE_SYNTAXE_ERROR;
	       return(NULL);
	     }
	   }
           if(!info_in_fill)
	   {info_begin = buf_pos; info_in_fill = TRUE; info_nothing_fill = FALSE;}
   	   buf_pos++;
	   break;
      }
    }
    g_free(buf);
    if(type_str)    g_free(type_str);
    if(program_str) g_free(program_str);
    if(icon_name)   g_free(icon_name);			    
  }

  return(vFileTypeList);
}
/******************************************************/

/******************************************************************/
/* add all needed file types to the file type list pFileTypeList. */
/* pFileTypeList MUST exist before this call.                     */
void file_type_list_add_needed_file_type(FileTypeList *pFileTypeList)
{ guint8   mode;  
  FileType *vFileType;
  
  /* check and create if needed the needed types */
  for (mode = NORMAL_DIRECTORY; mode < LAST_MUST_BE_TYPE; mode++)
  {
    if(file_type_list_find_file_type(pFileTypeList,mode,FALSE) == NULL)
    {
      vFileType = file_type_new(mode,NULL,NULL,NULL,FALSE,NULL,NULL);
      file_type_list_append_file_type(pFileTypeList,vFileType);
    }
    if(file_type_list_find_file_type(pFileTypeList,mode,TRUE) == NULL)
    {
      vFileType = file_type_new(mode,NULL,NULL,NULL,TRUE,NULL,NULL);
      file_type_list_append_file_type(pFileTypeList,vFileType);
    }
  }
}
/******************************************************************/

/*******************************************************************/
/* Try to found the icon "icon_name" in pFileTypeList. If found    */
/* pixmap and mask are filled with it corresponding pixmap & mask. */
/* Use this to avoid multiple loading a the same icon. pixmap &    */
/* mask are filled with NULL if icon is not found.                 */
void file_type_list_find_icon(FileTypeList *pFileTypeList, gchar *icon_name,
  GdkPixmap **pixmap, GdkBitmap **mask)
{
  *pixmap = NULL; *mask = NULL;
  
  if(icon_name == NULL) return;

  file_type_list_init(pFileTypeList);
  while(file_type_list_is_end(pFileTypeList) == FALSE)
  {
    if((file_type_list_current(pFileTypeList)->icon_name != NULL) &&
       (!strcmp(file_type_list_current(pFileTypeList)->icon_name,icon_name)))
    {
      *pixmap = gdk_pixmap_ref(file_type_list_current(pFileTypeList)->pixmap);
      *mask   = gdk_bitmap_ref(file_type_list_current(pFileTypeList)->mask);
      return;
    }
    file_type_list_next(pFileTypeList);
  }
}
/*******************************************************************/

/****************************************/
/* move FileType at pos src to pos dst. */
void file_type_list_move_file_type(FileTypeList *pFileTypeList,FileType *src,
  FileType *dst)
{
  gint pos_src = 0;
  gint pos_dst = 0;
  GSList *src_node = NULL;
  GSList *dst_node = NULL;
  GSList *src_node_past = NULL;
  GSList *dst_node_past = NULL;
  GSList *list_node_past = NULL;

  /* security check */
  if((pFileTypeList == NULL) || (src == NULL) || (dst == NULL) ||
     (src == dst)) return;

  file_type_list_init(pFileTypeList);
  list_node_past = pFileTypeList->list;
  while((file_type_list_is_end(pFileTypeList) == FALSE) && 
        !(src_node && dst_node) )
  {    
    if(file_type_list_current(pFileTypeList) == src) {
      src_node = pFileTypeList->current;
      src_node_past = list_node_past;
    } else if(file_type_list_current(pFileTypeList) == dst) {
      dst_node = pFileTypeList->current;
      dst_node_past = list_node_past;
    }

    if(src_node == NULL) pos_src++;
    if(dst_node == NULL) pos_dst++;
   
    list_node_past = pFileTypeList->current;
    file_type_list_next(pFileTypeList);
  }

  /* if one or both of types not found, exit */
  if((src_node == NULL) || (dst_node == NULL)) return;

  /* remove the src */
  src_node_past->next = src_node->next;

  /* insert src at the right pos */
  if(pos_src > pos_dst)
  {
    src_node->next = dst_node;
    dst_node_past->next = src_node;
  } else {
    src_node->next = dst_node->next;
    dst_node->next = src_node;
  }
}
/*******************************************************/
