/**********************************************/
/*  FileTypeList.c headers                    */
/*					      */
/*  Description : FileTypeList take care of   */
/*		  file type reconization list */
/*					      */
/*  Written by : Daniel Lacroix	              */
/*  GFile is under the GPL2 licence           */
/**********************************************/

#ifndef __FILE_TYPE_LIST_H__
#define __FILE_TYPE_LIST_H__

/* files bigger than that will be refused beeing 
   loaded by file_type_list_load_from_file */
#define FILE_TYPE_LIST_MAX_FILE_SIZE  500000

/* error number possible in gfile_errno */
#define FILE_TYPE_LIST_FILE_TOO_BIG       1
#define FILE_TYPE_LIST_FILE_OPEN_ERROR    2
#define FILE_TYPE_LIST_FILE_SYNTAXE_ERROR 3
#define FILE_TYPE_LIST_FILE_INVALID_FILE_TYPE 4
#define FILE_TYPE_LIST_FILE_ICON_NOT_FOUND    5

#define FILE_TYPE_LIST_CREAT_FILE_ERROR   6
#define FILE_TYPE_LIST_WRITE_FILE_ERROR   7

typedef struct _FileTypeList FileTypeList;

struct _FileTypeList
{
  /* the list of all the file type */
  GSList *list;
  /* the two above are for going over the list */
  /* current type */
  GSList *current;
};

/* Create a new file type system.          */
/* Return : the new allocated FileTypeList */
FileTypeList *file_type_list_new(void);

/* Free a file type system. */
void file_type_list_free(FileTypeList *pFileTypeList);

/*********************************************************/
/************ FOR GOING OVER ALL THE FILE TYPES **********/

/* After a call to this procedure, file_type_list_current */
/* return the first FileType                              */
#define file_type_list_init(pFileTypeList)    (pFileTypeList->current = g_slist_next(pFileTypeList->list))

/* Return the FileType we are on */
#define file_type_list_current(pFileTypeList) ((FileType *)pFileTypeList->current->data)

/* Go to the next FileType */
#define file_type_list_next(pFileTypeList)    (pFileTypeList->current = g_slist_next(pFileTypeList->current))

/* say if there is no more FileType */
#define file_type_list_is_end(pFileTypeList)  (pFileTypeList->current == NULL)

/*********************************************************/

/* insert after "before" in the list   */
/* if before = NULL, insert at the end */
void file_type_list_insert_file_type(FileTypeList *pFileTypeList,
  FileType *before, FileType *pfile_type);

/* add at the end of the list */
#define file_type_list_append_file_type(pFileTypeList,pFileType) file_type_list_insert_file_type(pFileTypeList,NULL,pFileType)

/* Remove a file type to the file type list     */
void file_type_list_remove_file_type(FileTypeList *pFileTypeList,
  FileType *remove_type);

/* swap two FileTypeList.                                        */
/* You CAN'T CALLED THIS IF YOU ARE USING file_type_list_next... */
void file_type_list_swap_file_type_list(FileTypeList *pFileTypeList1,
  FileTypeList *pFileTypeList2);

/* swap two file type in a file type list */
void file_type_list_swap_file_type(FileTypeList *pFileTypeList,FileType *type1,
  FileType *type2);

/* add all the file types of psrc at the end of pdest. */
/* If pdest is empty, this make a copy of psrc.        */
void file_type_list_append_file_type_list(FileTypeList *pdest,
  FileTypeList *psrc);

/* Return a copy of psrc file type list */
FileTypeList *file_type_list_copy(FileTypeList *psrc);

/* Return the corresponding USER_DEFINER type for    */
/* filename link or not. Return NULL if none defined */
FileType *file_type_list_find_user_defined_file_type(
  FileTypeList *pFileTypeList,gchar *filename,gboolean is_link);

/* Return the FileType corresponding to mode */
/* link or not or return NULL if not found   */
FileType *file_type_list_find_file_type(FileTypeList *pFileTypeList,
  guint8 mode,gboolean is_link);

/* Save the file type list in the filetype file */
/* filename is the full path name of the file   */
/* Return : FALSE = impossible; TRUE = job done */
/* if FALSE, gfile_errno can be set to          */
/*   FILE_TYPE_LIST_CREAT_FILE_ERROR            */
/*   FILE_TYPE_LIST_WRITE_FILE_ERROR            */
gboolean file_type_list_save_to_file(FileTypeList *pFileTypeList,gchar *filename);

/* Load the file type list from a file.               */
/* filename is the full path name of the file         */
/* Return : NULL = impossible gfile_errno can be set  */
/*          to FILE_TYPE_LIST_FILE_TOO_BIG            */
/*             FILE_TYPE_LIST_FILE_OPEN_ERROR         */
/*             FILE_TYPE_LIST_FILE_SYNTAXE_ERROR      */
/*            FILE_TYPE_LIST_FILE_INVALID_FILE_TYPE   */
/*            FILE_TYPE_LIST_FILE_ICON_NOT_FOUND      */
/*          else job done and the file type list fill */
FileTypeList *file_type_list_load_from_file(gchar *filename);

/* add all needed file types to the file type list pFileTypeList. */
/* pFileTypeList MUST exist before this call.                     */
void file_type_list_add_needed_file_type(FileTypeList *pFileTypeList);

/* Try to found the icon "icon_name" in pFileTypeList. If found    */
/* pixmap and mask are filled with it corresponding pixmap & mask. */
/* Use this to avoid multiple loading a the same icon. pixmap &    */
/* mask are filled with NULL if icon is not found.                 */
void file_type_list_find_icon(FileTypeList *pFileTypeList, gchar *icon_name,
  GdkPixmap **pixmap, GdkBitmap **mask);

/* move FileType at pos src to pos dst. */
void file_type_list_move_file_type(FileTypeList *pFileTypeList,FileType *src,
  FileType *dst);

#endif /* __FILE_TYPE_LIST_H__ */
