/****************************************/
/*  GFileTree.c body    	       	*/
/*					*/
/*  Description : Manage the tree list  */
/*		  widget		*/
/*					*/
/*  Written by : Daniel Lacroix	        */
/*  GFile is under the GPL2 licence     */
/****************************************/

#include "GFile.h"

/********************************************************************/
/* TO DO                                                            */
/* - drag & drop                                                    */
/* - hot preference changing                                        */
/* - support of pop_up_prop_menu                                    */
/* - One function to insert/supress a new item                      */
/* - make the sort working                                          */
/* - a function to open de tree until a given directory             */
/* - support for localization                                       */
/* - different icons for link on directory (perhaps)                */
/* - check the case when suppressing a dir we are on but with the   */
/*   list part active. Check if the active part if changing or not  */
/* - check all the case when the list is going in a directory and   */
/*   the tree can't follow.                                         */
/********************************************************************/

guint gfiletree_get_type ();
GtkWidget* gfiletree_new (GFileTreePref pGFileTreePref);
void gfiletree_free(GFileTree *pGFileTree);

void gfiletree_up(GFileTree *pGFileTree, gchar *directory);
void gfiletree_down(GFileTree *pGFileTree);

void gfiletree_update_filetype(GFileTree *pGFileTree,
  FileTypeList *new_file_type_list);
void gfiletree_add_dir_by_name(GFileTree *pGFileTree,gchar *directory);
void gfiletree_remove_dir_by_name(GFileTree *pGFileTree,gchar *directory);
void gfiletreepref_copy(GFileTreePref *pGFileTreePref1,
  GFileTreePref *pGFileTreePref2);

/* internal use only */
static void gfiletree_select_cb(GtkCTree *ctree_cb,GtkCTreeNode *row_cb,
                         gint column,GFileTree *pGFileTree);
static void gfiletree_expand_cb(GtkCTree *ctree_cb, GtkCTreeNode *node,
                         GFileTree *pGFileTree);
static gboolean gfiletree_load_path(GFileTree *pGFileTree,gchar *path,
                                  GtkCTreeNode *parent,guint8 nb_level);
static void string_marshal(GtkObject *object, GtkSignalFunc func, 
  gpointer func_data, GtkArg *args);
static void gfiletree_load_root_tree(GFileTree *pGFileTree);
static void gfiletree_update_filetype_part(GFileTree *pGFileTree,GtkCTreeNode *node);
static GtkCTreeNode *gfiletree_find_dir_at_node(GFileTree *pGFileTree,
                        GtkCTreeNode *parent,gchar *directory);
static void gfiletree_class_init(GFileTreeClass *class);
static void gfiletree_init (GFileTree *pGFileTree);
static GtkCTreeNode *gfiletree_find_path(GFileTree *pGFileTree, gchar *path);
static gint gfiletree_event_cb(GtkCList *clist, GdkEvent *event,
  GFileTree *pGFileTree);

/* if set to false, the change signal is never emit */
gboolean allow_change_signal = TRUE;

/* The default Preference */
GFileTreePref DefaultGFileTreePref =
{
  FALSE,        /* don't hide the title */
  "Directory",
  SORT_BY_NAME,	/* sort mode */
  TRUE,         /* TRUE = allow the menu where we can found  */
  FALSE,        /* TRUE = file starting with '.' are display */
  TRUE,         /* TRUE = show status bar                    */
  NULL          /* file_type_list can't have a default value */
};


/* for the drag&drop */
static void source_drag_data_get(GtkWidget *widget,GdkDragContext *context,
  GtkSelectionData *selection_data,guint info,guint time,GFileTree *pGFileTree);
  
enum { TARGET_FILES };
static GtkTargetEntry target_table[] = {
  { "text/uri-list", 0, TARGET_FILES }
};
static guint n_targets = sizeof(target_table) / sizeof(target_table[0]);

/*************************************************************/
guint gfiletree_get_type ()
{
  static guint gfiletree_type = 0;

  if (!gfiletree_type)
  {
    GtkTypeInfo gfiletree_info =
    {
       "GFileTree",
       sizeof (GFileTree),
       sizeof (GFileTreeClass),
       (GtkClassInitFunc)  gfiletree_class_init,
       (GtkObjectInitFunc) gfiletree_init,
       (GtkArgSetFunc) NULL,
       (GtkArgGetFunc) NULL
     };
     gfiletree_type = gtk_type_unique(gtk_vbox_get_type(), &gfiletree_info);
  }
  return gfiletree_type;
}
/*************************************************************/

/*************************************************************/
enum { CHANGE_DIR_SIGNAL, LAST_SIGNAL};
static gint gfiletree_signals[LAST_SIGNAL] = { 0 };

static void string_marshal(GtkObject *object, GtkSignalFunc func, gpointer func_data, GtkArg *args) {
    typedef void (*string_marshal_func)(GFileTree *, gchar *, gpointer);
    string_marshal_func rfunc = (string_marshal_func) func;
    (*rfunc) (GFILETREE(object),
	      GTK_VALUE_POINTER(args[0]), func_data);
}

static void gfiletree_class_init(GFileTreeClass *class)
{
/* here we create the new signals if we want */
  GtkObjectClass *object_class;
  object_class = (GtkObjectClass*) class;

  gfiletree_signals[CHANGE_DIR_SIGNAL] = gtk_signal_new ("change_dir",
            GTK_RUN_FIRST,object_class->type,
            GTK_SIGNAL_OFFSET (GFileTreeClass,change_dir),
            string_marshal, GTK_TYPE_NONE, 1,GTK_TYPE_POINTER);
	    
   gtk_object_class_add_signals (object_class, gfiletree_signals, LAST_SIGNAL);
   class->change_dir = NULL;
}
/*************************************************************/

/* copy the content of  pGFileTreePref2 in pGFileTreePref1 */
void gfiletreepref_copy(GFileTreePref *pGFileTreePref1,
  GFileTreePref *pGFileTreePref2)
{
  memcpy(pGFileTreePref1,pGFileTreePref2,sizeof(GFileTreePref));
/*  pGFileTreePref1->hide_title = pGFileTreePref2->hide_title;
*/
  pGFileTreePref1->title = g_strdup(pGFileTreePref2->title);
/*  
  pGFileTreePref1->sort_mode = pGFileTreePref2->sort_mode;
  pGFileTreePref1->file_type_list = pGFileTreePref2->file_type_list;
*/
}

/*********************/
/* set the sort mode */
static void gfiletree_set_sort_internal(GFileTree *pGFileTree, guint8 sort_mode)
{
  switch(sort_mode)
  {
    case SORT_NONE           :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_none_func); break;    
    case SORT_BY_NAME        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_name_func); break;    
    case SORT_BY_OWNER       :  
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_owner_func); break;    
    case SORT_BY_GROUP       :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_group_func); break;    
    case SORT_BY_RIGHTS      :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_rights_func); break;
    case SORT_BY_SIZE        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_size_func); break;    
    case SORT_BY_ACCESS_DATE :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_access_date_func); break;        
    case SORT_BY_MODIF_DATE  :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_modif_date_func); break;    
    case SORT_BY_CHANGE_DATE   :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_change_date_func); break;    

    case SORT_NONE | SORT_INVERT_ORDER           :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_none_inv_func); break;    
    case SORT_BY_NAME | SORT_INVERT_ORDER        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_name_inv_func); break;    
    case SORT_BY_OWNER | SORT_INVERT_ORDER       :  
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_owner_inv_func); break;    
    case SORT_BY_GROUP | SORT_INVERT_ORDER       :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_group_inv_func); break;    
    case SORT_BY_RIGHTS | SORT_INVERT_ORDER      :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_rights_inv_func); break;
    case SORT_BY_SIZE | SORT_INVERT_ORDER        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_size_inv_func); break;    
    case SORT_BY_ACCESS_DATE | SORT_INVERT_ORDER :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_access_date_inv_func); break;        
    case SORT_BY_MODIF_DATE | SORT_INVERT_ORDER  :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_modif_date_inv_func); break;    
    case SORT_BY_CHANGE_DATE | SORT_INVERT_ORDER   :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_change_date_inv_func); break;

    case SORT_NONE | SORT_DIR_FIRST          :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_none_dir_first_func); break;    
    case SORT_BY_NAME | SORT_DIR_FIRST        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_name_dir_first_func); break;    
    case SORT_BY_OWNER | SORT_DIR_FIRST       :  
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_owner_dir_first_func); break;    
    case SORT_BY_GROUP | SORT_DIR_FIRST       :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_group_dir_first_func); break;    
    case SORT_BY_RIGHTS | SORT_DIR_FIRST      :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_rights_dir_first_func); break;
    case SORT_BY_SIZE | SORT_DIR_FIRST        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_size_dir_first_func); break;    
    case SORT_BY_ACCESS_DATE | SORT_DIR_FIRST :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_access_date_dir_first_func); break;        
    case SORT_BY_MODIF_DATE | SORT_DIR_FIRST  :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_modif_date_dir_first_func); break;    
    case SORT_BY_CHANGE_DATE | SORT_DIR_FIRST   :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_change_date_dir_first_func); break;    

    case SORT_NONE | SORT_INVERT_ORDER | SORT_DIR_FIRST           :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_none_inv_dir_first_func); break;    
    case SORT_BY_NAME | SORT_INVERT_ORDER | SORT_DIR_FIRST        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_name_inv_dir_first_func); break;    
    case SORT_BY_OWNER | SORT_INVERT_ORDER | SORT_DIR_FIRST       :  
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_owner_inv_dir_first_func); break;    
    case SORT_BY_GROUP | SORT_INVERT_ORDER | SORT_DIR_FIRST       :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_group_inv_dir_first_func); break;    
    case SORT_BY_RIGHTS | SORT_INVERT_ORDER | SORT_DIR_FIRST      :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_rights_inv_dir_first_func); break;
    case SORT_BY_SIZE | SORT_INVERT_ORDER | SORT_DIR_FIRST        :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_size_inv_dir_first_func); break;    
    case SORT_BY_ACCESS_DATE | SORT_INVERT_ORDER | SORT_DIR_FIRST :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_access_date_inv_dir_first_func); break;        
    case SORT_BY_MODIF_DATE | SORT_INVERT_ORDER | SORT_DIR_FIRST  :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_modif_date_inv_dir_first_func); break;    
    case SORT_BY_CHANGE_DATE | SORT_INVERT_ORDER | SORT_DIR_FIRST   :
      gtk_clist_set_compare_func(GTK_CLIST(pGFileTree->Ctree),
  	sort_by_change_date_inv_dir_first_func); break;	    
  }
}

void gfiletree_set_sort(GFileTree *pGFileTree, guint8 sort_mode)
{
  gfiletree_set_sort_internal(pGFileTree, sort_mode);
  
  pGFileTree->Preference.sort_mode = sort_mode;
  gtk_ctree_sort_recursive(GTK_CTREE(pGFileTree->Ctree),pGFileTree->root_node);
}
/*********************/

/*******************************/
/* Update de status bar if set */
gfiletree_fill_status(GFileTree *pGFileTree)
{ struct statfs buf;
  gchar *to_display;

  if(pGFileTree->statusbar == FALSE) return;
  
  statfs(pGFileTree->current_path, &buf);

  if((buf.f_blocks>>10)*buf.f_bsize < 10000)
    to_display = g_strdup_printf(" %d / %d Ko (free %d Ko)",
      ((buf.f_blocks - buf.f_bfree)*buf.f_bsize)>>10,
      (buf.f_blocks*buf.f_bsize)>>10,
      (buf.f_bavail*buf.f_bsize)>>10);
  else
    to_display = g_strdup_printf(" %d / %d Mo (free %d Mo)",
      (((buf.f_blocks - buf.f_bfree)>>10)*buf.f_bsize)>>10,
      ((buf.f_blocks>>10)*buf.f_bsize)>>10,
      ((buf.f_bavail>>10)*buf.f_bsize)>>10);
  
  if ((GTK_STATUSBAR(pGFileTree->statusbar))->messages)
     gtk_statusbar_pop (GTK_STATUSBAR(pGFileTree->statusbar), 1);
  gtk_statusbar_push (GTK_STATUSBAR(pGFileTree->statusbar), 1,to_display);

  g_free(to_display);
}
/*******************************/

/* use as a parameter given by gfiletree_new to gfiletree_init */
GFileTreePref current_GFileTreePref;

/*************************************************************/
void gfiletree_destroy_cb(GFileTree *pGFileTree, gpointer data);

static void gfiletree_init (GFileTree *pGFileTree)
{
  GtkWidget *scrolled_widget;
  gchar     *title[1];
  
  /* take parameters in account */
  gfiletreepref_copy((GFileTreePref *)(&(pGFileTree->Preference)),
    &current_GFileTreePref);
  g_free(current_GFileTreePref.title);

  title[0] = pGFileTree->Preference.title;
  pGFileTree->Ctree = GTK_CTREE(gtk_ctree_new_with_titles(1,0,title));

  if(pGFileTree->Preference.hide_title == TRUE) 
    gtk_clist_column_titles_hide(GTK_CLIST(pGFileTree->Ctree));
    
  gtk_ctree_set_line_style (pGFileTree->Ctree, GTK_CTREE_LINES_DOTTED);
  gtk_widget_show(GTK_WIDGET(pGFileTree->Ctree));

  /* when the widget is destroy, called gfiletree_free */
  gtk_signal_connect(GTK_OBJECT(pGFileTree),"destroy",
		     GTK_SIGNAL_FUNC(gfiletree_destroy_cb),pGFileTree);

  gtk_signal_connect_after(GTK_OBJECT(pGFileTree->Ctree),"tree_select_row",
			   GTK_SIGNAL_FUNC(gfiletree_select_cb),pGFileTree);

  gtk_signal_connect_after(GTK_OBJECT(pGFileTree->Ctree),"tree_expand",
			   GTK_SIGNAL_FUNC (gfiletree_expand_cb),pGFileTree);

  gtk_signal_connect(GTK_OBJECT(pGFileTree->Ctree),"event",
    GTK_SIGNAL_FUNC(gfiletree_event_cb), pGFileTree);

  /* define the scroll properties */
  scrolled_widget = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrolled_widget),
    GTK_POLICY_AUTOMATIC,GTK_POLICY_ALWAYS);
  gtk_widget_show(scrolled_widget);
  gtk_container_add(GTK_CONTAINER(scrolled_widget),GTK_WIDGET(pGFileTree->Ctree));

  gtk_box_pack_start(GTK_BOX(pGFileTree),GTK_WIDGET(scrolled_widget),
                     TRUE,TRUE,0);

  gfiletree_set_sort_internal(pGFileTree, pGFileTree->Preference.sort_mode);
  /* load the root content */		     
  pGFileTree->current_path = NULL;
  gfiletree_load_root_tree(pGFileTree);
  
  if(pGFileTree->Preference.statusbar == TRUE)
  {
    pGFileTree->statusbar = gtk_statusbar_new();
    gtk_box_pack_start(GTK_BOX(pGFileTree),pGFileTree->statusbar,FALSE,TRUE,0);
    gtk_widget_show(pGFileTree->statusbar);
  }
  gfiletree_fill_status(pGFileTree);

  /* over here this is for drag&drop */
  /* set the list as a source */
  gtk_drag_source_set(GTK_WIDGET(pGFileTree->Ctree),GDK_BUTTON1_MASK|GDK_BUTTON3_MASK,
    target_table, n_targets, GDK_ACTION_COPY | GDK_ACTION_MOVE);

  gtk_signal_connect(GTK_OBJECT(pGFileTree->Ctree),"drag_data_get",
    GTK_SIGNAL_FUNC(source_drag_data_get),pGFileTree);

}
/*************************************************************/

/*************************************************************/
GtkWidget* gfiletree_new (GFileTreePref pGFileTreePref)
{
  gfiletreepref_copy(&current_GFileTreePref,&pGFileTreePref);
  return GTK_WIDGET ( gtk_type_new( gfiletree_get_type() ) );
}
/*************************************************************/

/*************************************************/
/* Free a GFileTree				 */
static void free_node_data(GtkCTree *ctree,GtkCTreeNode *node,gpointer data)
{ FileData *vFileData;
 
  vFileData = gtk_ctree_node_get_row_data(ctree,node);
  /*g_print("directory : >%s%s<\n",vFileData->path,vFileData->name);*/
  file_data_free(vFileData);
}

void gfiletree_destroy_cb(GFileTree *pGFileTree, gpointer data)
{
  /*g_print("A GFileTree will be destroy\n");*/
  /* free the current path */
  g_free(pGFileTree->current_path);
  /* free the title */
  g_free(pGFileTree->Preference.title);
  /* free all the ctree */
  gtk_ctree_post_recursive(GTK_CTREE(pGFileTree->Ctree),
    GTK_CTREE_NODE(pGFileTree->root_node),GTK_CTREE_FUNC(free_node_data),NULL);
}
/*************************************************/

/********************************************************/
/* Return TRUE if this FileData is accepted in the tree */
/* (directory only...)                                  */
gboolean gfiletree_filter(GFileTree *pGFileTree, FileData *pFileData)
{
         /* it must be different than . and .. */
  return(strcmp(pFileData->name,".") && strcmp(pFileData->name,"..") &&
         /* it must be a directory or a link on a directory */
         (S_ISDIR(pFileData->stat.st_mode) || (S_ISLNK(pFileData->stat.st_mode) && 
          (!(pFileData->extra_info & FILE_DATA_IS_BROKEN_SYMB)) && 
          S_ISDIR(pFileData->link_stat->st_mode))) &&
	 /* it must start with something different than . if we hide hidden file */
	 !((pFileData->name[0] == '.') && !pGFileTree->Preference.show_hidden_file)
	);
}
/********************************************************/

/***************************************************/
/* Attention : DON'T ENDED by path by a '/'        */
/*               (except for '/' directory)        */
/* Return : FALSE if this directory doesn't exist  */
/*          or parent are bad. Else : TRUE         */
typedef struct
{
  GFileTree *vGFileTree;
  FileType  *normal_dir_type;
  FileType  *normal_open_dir_type;
  guint8    nb_level;
  GtkCTreeNode *parent;
} user_func_info;

static gboolean user_func(FileData *pFileData,gpointer data)
{ gchar *text[1];
  GtkCTreeNode *sibling;
  user_func_info *vuser_func_info;
  gchar *directory;

  vuser_func_info = (user_func_info *)data;

  if(gfiletree_filter(vuser_func_info->vGFileTree, pFileData))
  {
    text[0] = pFileData->name;
    sibling = gtk_ctree_insert_node(vuser_func_info->vGFileTree->Ctree,
      vuser_func_info->parent,NULL,text,5,
      vuser_func_info->normal_dir_type->pixmap,
      vuser_func_info->normal_dir_type->mask,
      vuser_func_info->normal_open_dir_type->pixmap,
      vuser_func_info->normal_open_dir_type->mask, FALSE,FALSE);
    gtk_ctree_node_set_row_data(vuser_func_info->vGFileTree->Ctree,sibling,pFileData);

    if(vuser_func_info->nb_level) /* scan nb_level under level in the tree */
    {
      directory = g_strconcat(pFileData->path,pFileData->name,NULL);
      gfiletree_load_path(vuser_func_info->vGFileTree,directory,sibling,vuser_func_info->nb_level-1);
      g_free(directory);
    }
    return(TRUE);
  } else
    return(FALSE);
}

static gboolean gfiletree_load_path(GFileTree *pGFileTree,gchar *path,
                             GtkCTreeNode *parent,guint8 nb_level)
{
  user_func_info vuser_func_info;

  /* a protection */
  if((path == NULL) || (parent == NULL) || (pGFileTree==NULL)) return(FALSE);

  /* in the tree, no difference are made between link to dir or dir */
  vuser_func_info.normal_dir_type = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    NORMAL_DIRECTORY,FALSE);
  if(vuser_func_info.normal_dir_type == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  vuser_func_info.normal_open_dir_type = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    NORMAL_OPEN_DIRECTORY,FALSE);
  if(vuser_func_info.normal_open_dir_type == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}

  vuser_func_info.vGFileTree = pGFileTree;
  vuser_func_info.nb_level = nb_level;
  vuser_func_info.parent = parent;

  if(file_data_open_dir(path,pGFileTree->Preference.file_type_list,NULL,
    user_func,&vuser_func_info) == FALSE) return(FALSE);
  
  /* sort the tree DOESN'T WORK (WHY ?)*/
  gtk_ctree_sort_recursive(GTK_CTREE(pGFileTree->Ctree),parent);
  return(TRUE);
}
/***************************************************/

/***************************************************/
/* Load the root tree. This must be done only when */
/* we create a new GFileTree.			   */
static void gfiletree_load_root_tree(GFileTree *pGFileTree)
{
  FileType *tmptype, *tmptype2;

  gchar *message[1]={"/"};
  gchar *text[1];
  GtkCTreeNode *sibling;
  FileData *vFileData;
  
  /* needed to forced root_node to be in current mode */
  gchar *tmp_text;
  guint8 spacing;
  GdkPixmap *pixmap_cb;
  GdkBitmap *mask_cb;
  gboolean is_leaf;
  gboolean expanded;
  GdkPixmap *pixmap_info;
  GdkBitmap *mask_info;
  gchar *tmp_string;

  gtk_clist_freeze(GTK_CLIST(pGFileTree->Ctree));
  /* clear to be sure there is nothing */
  gtk_clist_clear(GTK_CLIST(pGFileTree->Ctree));
  
  /* find directory types */
  tmptype = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    PAST_DIRECTORY,FALSE);
  if(tmptype == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  tmptype2 = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    PAST_OPEN_DIRECTORY,FALSE);
  if(tmptype2 == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  
  /* place the root_node */
  pGFileTree->root_node=gtk_ctree_insert_node(pGFileTree->Ctree,
  		NULL, NULL, message, 5, tmptype->pixmap,
		tmptype->mask, tmptype2->pixmap, tmptype2->mask,
		FALSE, TRUE);

  vFileData = g_malloc(sizeof(FileData));
  vFileData->path = g_strdup("");
  vFileData->name = g_strdup("/");
  vFileData->type = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    NORMAL_DIRECTORY,FALSE);
  if(vFileData->type == NULL)
  {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  vFileData->link_to_file = NULL;
  vFileData->extra_info = FILE_DATA_IS_DIR_SCAN;
  vFileData->link_stat = NULL;
  vFileData->ref_count = 1;

  gtk_ctree_node_set_row_data(pGFileTree->Ctree,pGFileTree->root_node,
                              vFileData);
  /* scan at 2 sub levels */
  gfiletree_load_path(pGFileTree,"/",pGFileTree->root_node,1);
  gtk_clist_thaw(GTK_CLIST(pGFileTree->Ctree));

  /* fill the path */
  pGFileTree->last_node = pGFileTree->root_node;
  if(pGFileTree->current_path != NULL) g_free(pGFileTree->current_path);
  pGFileTree->current_path = g_strdup("/");

  /* select the root_node */
  gtk_signal_emit_by_name(GTK_OBJECT(pGFileTree->Ctree),
             "tree_select_row",pGFileTree->root_node,0);

  /* we must forced the root to be in "current mode" */
  gtk_ctree_get_node_info(pGFileTree->Ctree,pGFileTree->root_node,&tmp_text,
    &spacing,&pixmap_info,&mask_info,&pixmap_cb,&mask_cb,&is_leaf,&expanded);
  tmptype = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    CURRENT_DIRECTORY,FALSE);
  if(tmptype == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  tmptype2 = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    CURRENT_OPEN_DIRECTORY,FALSE);    
  if(tmptype2 == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
 
  tmp_string = g_strdup(tmp_text);
  /* change to have the style of the current file type directory */
  gtk_ctree_set_node_info(pGFileTree->Ctree,pGFileTree->root_node,tmp_string,
    spacing,tmptype->pixmap,tmptype->mask, tmptype2->pixmap,tmptype2->mask,
    is_leaf,expanded);
  g_free(tmp_string);
}
/***************************************************/

/*******************************************/
/* called each time a tree row is selected */
static void gfiletree_select_cb(GtkCTree *ctree_cb, GtkCTreeNode *row_cb, 
		    gint column,GFileTree *pGFileTree)
{ gchar *text;
  guint8 spacing;
  GdkPixmap *pixmap_cb;
  GdkBitmap *mask_cb;
  GtkCTreeNode *last_node_cb;
  gboolean is_leaf;
  gboolean expanded;
  GdkPixmap *pixmap_info;
  GdkBitmap *mask_info;
  FileType *tmptype, *tmptype2;

  FileData  *vFileData;
  FileData  *vFileData2;
  gchar *tmp_string;

  /* freeze the tree to avoid fliker */
  gtk_clist_freeze(GTK_CLIST(pGFileTree->Ctree));

  /* do if the guy click on something new */
  if(row_cb != pGFileTree->last_node)
  {
    gtk_ctree_node_get_pixtext(ctree_cb,row_cb,column,&text,
                            &spacing,&pixmap_cb,&mask_cb);
    vFileData = gtk_ctree_node_get_row_data(ctree_cb,row_cb);
    /*g_print("ctree row selected signal : %s%s\n",vFileData->path,vFileData->name);*/
    /* change the current path */
    g_free(pGFileTree->current_path);
    /* put the new current_path */
    if(!strcmp(vFileData->name,"/"))
    {
      pGFileTree->current_path = g_strdup("/");
    } else {
      pGFileTree->current_path = g_strconcat(vFileData->path,vFileData->name,
        "/",NULL);
    }

    /* define last current node as a past directory */        
    gtk_ctree_get_node_info(pGFileTree->Ctree,pGFileTree->last_node,&text,&spacing,
                            &pixmap_info,&mask_info,&pixmap_cb,&mask_cb,
                            &is_leaf,&expanded);
    tmptype = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
      PAST_DIRECTORY,FALSE);
    if(tmptype == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
    tmptype2 = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
      PAST_OPEN_DIRECTORY,FALSE);
    if(tmptype2 == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}

    tmp_string = g_strdup(text);

    /* change to previous current file type directory into a past one */
    gtk_ctree_set_node_info(pGFileTree->Ctree,pGFileTree->last_node,tmp_string,
      spacing,tmptype->pixmap,tmptype->mask,
      tmptype2->pixmap,tmptype2->mask, is_leaf, expanded);

    g_free(tmp_string);

    /* define the current selected row as the current directory */
    gtk_ctree_get_node_info(ctree_cb,row_cb,&text,&spacing,
                            &pixmap_info,&mask_info,&pixmap_cb,&mask_cb,
                            &is_leaf,&expanded);
    tmptype = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
      CURRENT_DIRECTORY,FALSE);
    if(tmptype == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
    tmptype2 = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
      CURRENT_OPEN_DIRECTORY,FALSE);    
    if(tmptype2 == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}

    tmp_string = g_strdup(text);
    /* change to have the style of the current file type directory */
    gtk_ctree_set_node_info(ctree_cb,row_cb,tmp_string,spacing,
      tmptype->pixmap,tmptype->mask, tmptype2->pixmap,tmptype2->mask,
      is_leaf,expanded);
    g_free(tmp_string);

    pGFileTree->last_node = row_cb;

    /* emit the corresponding signal */
    if(allow_change_signal == TRUE)
    {
      tmp_string = g_strdup(pGFileTree->current_path);
      /* if current path is not only "/", suppress the last '/' */
      if(tmp_string[1] != 0) tmp_string[strlen(pGFileTree->current_path)-1] = 0;
      gtk_signal_emit(GTK_OBJECT(pGFileTree),
        gfiletree_signals[CHANGE_DIR_SIGNAL],tmp_string);
      g_free(tmp_string);
    }  

    /* scan the under tree if not known */
    if(!(vFileData->extra_info & FILE_DATA_IS_DIR_SCAN))
    {
      /*g_print("never scan\n");*/
      last_node_cb = GTK_CTREE_ROW(row_cb)->children;

      while((last_node_cb != NULL) && (GTK_CTREE_ROW(last_node_cb )->parent == row_cb))
      {
         vFileData2 = gtk_ctree_node_get_row_data(ctree_cb,last_node_cb);
	 tmp_string = g_strconcat(vFileData2->path,vFileData2->name,NULL);
         /*g_print("sous rep : %s\n",tmp_string);*/
	 gfiletree_load_path(pGFileTree,tmp_string,
                             last_node_cb,0);
	 g_free(tmp_string);
         last_node_cb = GTK_CTREE_NODE_NEXT(last_node_cb);
      }
      /* current directory is now scanned */
      vFileData->extra_info |= FILE_DATA_IS_DIR_SCAN;
    } /* end of if(!(vFileDa... */ /*else {g_print("already scan\n");}*/
  }

  gfiletree_fill_status(pGFileTree);

  /* Job finish. Unfreeze the tree. */
  gtk_clist_thaw (GTK_CLIST(pGFileTree->Ctree));
}
/*******************************************/

/************************************************************/
/* use this if you want to unselect every thing in the tree */
void gfiletree_unselect_all(GFileTree *pGFileTree)
{
  gtk_signal_emit_by_name(GTK_OBJECT(pGFileTree->Ctree),"tree_unselect_row",
    pGFileTree->last_node,0);
}
/************************************************************/

/*************************************************/
/* called each time a part of the tree is expand */
static void gfiletree_expand_cb(GtkCTree *ctree_cb, GtkCTreeNode *node,GFileTree *pGFileTree)
{
  GtkCTreeNode *last_node_cb;

  FileData  *vFileData;
  FileData  *vFileData2;
  gchar *tmp_string;

  /* freeze the tree to avoid fliker */
  gtk_clist_freeze(GTK_CLIST(pGFileTree->Ctree));

  vFileData = gtk_ctree_node_get_row_data(ctree_cb,node);
 
  /* scan the under tree if not known */
  if(!(vFileData->extra_info & FILE_DATA_IS_DIR_SCAN))
  {
    last_node_cb = GTK_CTREE_ROW(node)->children;

    while((last_node_cb != NULL) && (GTK_CTREE_ROW(last_node_cb)->parent == node))
    {
       vFileData2 = gtk_ctree_node_get_row_data(ctree_cb,last_node_cb);
       tmp_string = g_strconcat(vFileData2->path,vFileData2->name,NULL);
       /*g_print("sous rep : %s\n",tmp_string);*/
       gfiletree_load_path(pGFileTree,tmp_string,last_node_cb,0);
       g_free(tmp_string);
       last_node_cb = GTK_CTREE_NODE_NEXT(last_node_cb);
    }
    /* current directory is now scanned */
    vFileData->extra_info |= FILE_DATA_IS_DIR_SCAN;
  } /* end of if(!(vFileDa... */ /*else {g_print("already scan\n");}*/

  /* Job finish. Unfreeze the tree. */
  gtk_clist_thaw (GTK_CLIST(pGFileTree->Ctree));
  
}
/*************************************************/

/****************************************/
/* move up to directory in current path */
void gfiletree_up(GFileTree *pGFileTree, gchar *directory)
{ gchar *text;
  guint8 spacing;
  GdkPixmap *pixmap_cb;
  GdkBitmap *mask_cb;
  GtkCTreeNode *try_node;

  /*g_print("string to found : %s\n",directory);*/
  /* expand the tree at the current node */
  gtk_ctree_expand(pGFileTree->Ctree,pGFileTree->last_node);
  try_node = GTK_CTREE_ROW(pGFileTree->last_node)->children;
  /* try to found directory */
  
  while(((try_node)&&(GTK_CTREE_ROW(try_node)->parent==pGFileTree->last_node))||
        gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),pGFileTree->last_node,try_node))
  {

    while((!((try_node)&&(GTK_CTREE_ROW(try_node)->parent == pGFileTree->last_node))) &&
          gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),pGFileTree->last_node,try_node))
    { try_node = GTK_CTREE_NODE_NEXT(try_node); }

    gtk_ctree_node_get_pixtext(pGFileTree->Ctree,try_node,0,&text,
      &spacing,&pixmap_cb,&mask_cb);
    /*g_print("text choice : %s\n",text);*/
      
    if(!strcmp(directory,text))
    { 
      /*g_print("list expand found : %s\n",text);*/
      /* block the up_dir signal because the user already know */
      allow_change_signal = FALSE;
      gfiletree_select_cb(GTK_CTREE(pGFileTree->Ctree), try_node, 0, pGFileTree);
      allow_change_signal = TRUE;
      pGFileTree->last_node=try_node;
      return;
    }
    try_node = GTK_CTREE_NODE_NEXT(try_node);
  }
  /* Si on arrive jusque ici, c'est que l'arbre n'est plus bon
     par rapport a la realite. Ce cas n'est pas encore fait. */
  g_print("Erreur list : expand ctree\n");
}
/****************************************/

/****************************/
/* down to parent directory */
void gfiletree_down(GFileTree *pGFileTree)
{ GtkCTreeNode *new_node;

  if(pGFileTree->last_node != pGFileTree->root_node )
  {
    new_node = GTK_CTREE_ROW(pGFileTree->last_node)->parent;
    /* if we are not at the root */
    if(new_node != pGFileTree->root_node)
      gtk_ctree_collapse(pGFileTree->Ctree,new_node);
    /* block the up_dir signal because the user already know */
    allow_change_signal = FALSE;
    gfiletree_select_cb(GTK_CTREE(pGFileTree->Ctree), new_node, 0, pGFileTree);
    allow_change_signal = TRUE;
  }
}
/****************************/

/***************************************************/
/* This must be called when file type have changed */

/* The file type must have been change in pGFileTree->Preference.file_type_list */
static void gfiletree_update_filetype_part(GFileTree *pGFileTree,GtkCTreeNode *node)
{ GtkCTreeNode *vnode;
  FileType *type_close, *type_open;
  gchar *text,*text_copy;
  guint8 spacing;
  GdkPixmap   *pixmap_closed;
  GdkBitmap   *mask_closed;
  GdkPixmap   *pixmap_opened;
  GdkBitmap   *mask_opened;
  gboolean    is_leaf;
  gboolean    expanded;
  
  FileData    *vFileData;

  vnode = node;
  while(vnode != NULL)
  {
    if(GTK_CTREE_ROW(vnode)->children != NULL)
    {
      gfiletree_update_filetype_part(pGFileTree,GTK_CTREE_ROW(vnode)->children);
    }
    vFileData = gtk_ctree_node_get_row_data(pGFileTree->Ctree,vnode);
    if(vnode == pGFileTree->last_node)
    {
      type_close = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
        CURRENT_DIRECTORY,FALSE);
      if(type_close == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
      type_open = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
        CURRENT_OPEN_DIRECTORY,FALSE);
      if(type_open == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
    } else
    if(vFileData->extra_info & FILE_DATA_IS_DIR_SCAN)
    { 
      type_close = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
        PAST_DIRECTORY,FALSE);
      if(type_close == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
      type_open = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
        PAST_OPEN_DIRECTORY,FALSE);
      if(type_open == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
    } else {
      type_close = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
        NORMAL_DIRECTORY,FALSE);
      if(type_close == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
      type_open = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
        NORMAL_OPEN_DIRECTORY,FALSE);
      if(type_open == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
    }
    gtk_ctree_get_node_info(pGFileTree->Ctree,vnode,&text,&spacing,
 			    &pixmap_closed,&mask_closed,&pixmap_opened,
			    &mask_opened,&is_leaf,&expanded);
    text_copy = g_strdup(text);
    gtk_ctree_set_node_info(pGFileTree->Ctree,vnode,text_copy,spacing,
      type_close->pixmap,type_close->mask,type_open->pixmap,
      type_open->mask,is_leaf,expanded);
      
    g_free(text_copy);
    vnode = GTK_CTREE_NODE_NEXT(vnode);
  }
}

/* The caller is responsible for freeing the       */
/* previous file type list                         */
void gfiletree_set_filetype(GFileTree *pGFileTree,FileTypeList *new_file_type_list)
{
  /* Freeze the ctree to avoid flicker */
  gtk_clist_freeze(GTK_CLIST(pGFileTree->Ctree));
  pGFileTree->Preference.file_type_list = new_file_type_list;
  gfiletree_update_filetype_part(pGFileTree,pGFileTree->root_node);
  /* Unfreeze the ctree */
  gtk_clist_thaw(GTK_CLIST(pGFileTree->Ctree));  
}
/***************************************************/

/******************************************/
/* Find the node of directory (ex : usr)  */
/* in the list of child node of node      */
/* Return NULL if not found               */
static GtkCTreeNode *gfiletree_find_dir_at_node(GFileTree *pGFileTree,
                        GtkCTreeNode *parent,gchar *directory)
{ gchar *text;
  guint8 spacing;
  GdkPixmap *pixmap_cb;
  GdkBitmap *mask_cb;
  GtkCTreeNode *try_node;

  try_node = GTK_CTREE_ROW(parent)->children;
  while(((try_node)&&(GTK_CTREE_ROW(try_node)->parent==parent))||
        gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),parent,try_node))
  {

    while((!((try_node)&&(GTK_CTREE_ROW(try_node)->parent == parent))) &&
          gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),parent,try_node))
    { try_node = GTK_CTREE_NODE_NEXT(try_node); }

    gtk_ctree_node_get_pixtext(pGFileTree->Ctree,try_node,0,&text,
                            &spacing,&pixmap_cb,&mask_cb);
    /* if equal then we have found the node */
    if(!strcmp(directory,text)) return(try_node);

    try_node = GTK_CTREE_NODE_NEXT(try_node);
  }
  return(NULL);
}
/******************************************/

/*****************************************/
/* Add directory "directory" to the tree */
/* ex : directory=/usr/bin/newdir to add */
/* Do not end by a '/'                   */
extern void abs_2_rel(gchar *full_name,gchar **path,gchar **name);
void gfiletree_add_dir_by_name(GFileTree *pGFileTree,gchar *directory)
{
  FileType *normal_dir_type,*normal_open_dir_type;
  FileData *vFileData;
  GtkCTreeNode *parent,*sibling;
  gchar *path,*name;
  gboolean under_scan_needed = FALSE;
  gchar *error_msg;
  GtkWidget *error_dialog;

  /* separate path and name */
  abs_2_rel(directory,&path,&name);

  /* find the parent to insert */
  if((parent = gfiletree_find_path(pGFileTree,path)) == NULL)
  { g_free(path); g_free(name); return; }

  /* look if an under scan will be needed */
  vFileData = gtk_ctree_node_get_row_data(pGFileTree->Ctree,parent);
  if(vFileData->extra_info & FILE_DATA_IS_DIR_SCAN) {
    under_scan_needed = TRUE;
  } else if(GTK_CTREE_ROW(parent)->parent) {
    vFileData = gtk_ctree_node_get_row_data(GTK_CTREE(pGFileTree->Ctree),
      GTK_CTREE_ROW(parent)->parent);
    /* if the part we want to add something was never scan, forget about it */
    if(!(vFileData->extra_info & FILE_DATA_IS_DIR_SCAN))
    { g_free(path); g_free(name); return; }
  }

  vFileData = file_data_new_from_file(path,name,
    pGFileTree->Preference.file_type_list,NULL);
  /* ERROR message if needed */
  if(vFileData == NULL)
  {
    switch(gfile_errno){
      case FILE_DATA_LSTAT_ERROR :
        error_msg = g_strdup_printf("ERROR while stating %s",directory);
        break;
      default :
        error_msg = g_strdup("Unknown ERROR");
    }
    error_dialog = gnome_message_box_new(error_msg,GNOME_MESSAGE_BOX_ERROR,
      GNOME_STOCK_BUTTON_OK,NULL);
    g_free(error_msg);
    gnome_dialog_run_and_close(GNOME_DIALOG(error_dialog));
    /* and stop here of course */
    g_free(path); g_free(name);
    return;
  }

  /* if we don't accept those directory, file... exit */
  if(gfiletree_filter(pGFileTree, vFileData) == FALSE)
  { file_data_free(vFileData); g_free(path); g_free(name); return; }

  /* if the directory is already here, suppress it */
  if((sibling = gfiletree_find_dir_at_node(pGFileTree,parent,name)) != NULL)
  {
    gtk_ctree_post_recursive(GTK_CTREE(pGFileTree->Ctree), 
      GTK_CTREE_NODE(sibling),GTK_CTREE_FUNC(free_node_data),NULL);
    gtk_ctree_remove_node(GTK_CTREE(pGFileTree->Ctree),GTK_CTREE_NODE(sibling));
  }

  /* find the corresponding type to a directory */
  normal_dir_type = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    NORMAL_DIRECTORY,FALSE);
  if(normal_dir_type == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  normal_open_dir_type = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    NORMAL_OPEN_DIRECTORY,FALSE);
  if(normal_open_dir_type == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}

  sibling = gtk_ctree_insert_node(pGFileTree->Ctree,parent,NULL,&name,5,
    normal_dir_type->pixmap,normal_dir_type->mask,normal_open_dir_type->pixmap,
    normal_open_dir_type->mask, FALSE,FALSE);

  /* attach data */
  gtk_ctree_node_set_row_data(GTK_CTREE(pGFileTree->Ctree),sibling,vFileData);

  /* under scan if needed */
  if(under_scan_needed)
  { gfiletree_load_path(pGFileTree,directory,sibling,0); }

  g_free(path); g_free(name);
}
/*****************************************/

/************************************************************/
/* Remove the directory "directory" of the tree if it exist */
/* ex : directory=/usr/bin/newdir to del 		    */
/* You can end by a '/' or not.               		    */
void gfiletree_remove_dir_by_name(GFileTree *pGFileTree,gchar *directory)
{
  GtkCTreeNode *sibling;
  gchar *tmp_string;

  /* a protection */
  if((directory == NULL) || (directory[0]!='/')) return;

  /* if found, suppress it */
  if((sibling = gfiletree_find_path(pGFileTree,directory)) != NULL)
  {
    /* if we suppress the directory we are on, we must first go to
       the parent directory */
    if(sibling == pGFileTree->last_node)
    {
      gtk_signal_emit_by_name(GTK_OBJECT(pGFileTree->Ctree),
        "tree_select_row",GTK_CTREE_ROW(pGFileTree->last_node)->parent,0);

      tmp_string = g_strdup(pGFileTree->current_path);
      /* if current path is not only "/", suppress the last '/' */
      if(tmp_string[1] != 0) tmp_string[strlen(pGFileTree->current_path)-1] = 0;
      /* emit the signal to let the other know that the current directory has changed */
      gtk_signal_emit(GTK_OBJECT(pGFileTree),
        gfiletree_signals[CHANGE_DIR_SIGNAL],tmp_string);
      g_free(tmp_string);
    }
    /* now suppress safely */
    gtk_ctree_post_recursive(GTK_CTREE(pGFileTree->Ctree), 
      GTK_CTREE_NODE(sibling),GTK_CTREE_FUNC(free_node_data),NULL);
    gtk_ctree_remove_node(GTK_CTREE(pGFileTree->Ctree),GTK_CTREE_NODE(sibling));
  }
}
/************************************************************/

/************************************************************/
/* Modify the directory "directory" of the tree if it exist */
/* ex : directory=/usr/bin/newdir to modify 		    */
/* Do not end by a '/'                   		    */
void gfiletree_modify_dir_by_name(GFileTree *pGFileTree,gchar *directory,gchar *new_dir)
{
  GtkCTreeNode *sibling;
  FileData *vFileData;
  gchar *path,*name;
  gchar *error_msg;
  GtkWidget *error_dialog;

  gchar     *text;
  guint8    spacing;
  GdkPixmap *pixmap_closed;
  GdkBitmap *mask_closed;
  GdkPixmap *pixmap_opened;
  GdkBitmap *mask_opened;
  gboolean  is_leaf;
  gboolean  expanded;
  
  /* find the old directory */
  if((sibling = gfiletree_find_path(pGFileTree,directory)) == NULL)
  { return; }

  /* separate path and name */
  abs_2_rel(new_dir,&path,&name);

  vFileData = file_data_new_from_file(path,name,
    pGFileTree->Preference.file_type_list,NULL);
  /* ERROR message if needed */
  if(vFileData == NULL)
  {
    switch(gfile_errno){
      case FILE_DATA_LSTAT_ERROR :
        error_msg = g_strdup_printf("ERROR while stating %s",new_dir);
        break;
      default :
        error_msg = g_strdup("Unknown ERROR");
    }
    error_dialog = gnome_message_box_new(error_msg,GNOME_MESSAGE_BOX_ERROR,
      GNOME_STOCK_BUTTON_OK,NULL);
    g_free(error_msg);
    gnome_dialog_run_and_close(GNOME_DIALOG(error_dialog));
    /* and stop here of course */
    g_free(path); g_free(name);
    return;
  }

  /* if we don't accept those directory, file... exit and remove old one */
  if(gfiletree_filter(pGFileTree, vFileData) == FALSE)
  {
    gtk_ctree_post_recursive(GTK_CTREE(pGFileTree->Ctree), 
      GTK_CTREE_NODE(sibling),GTK_CTREE_FUNC(free_node_data),NULL);
    gtk_ctree_remove_node(GTK_CTREE(pGFileTree->Ctree),GTK_CTREE_NODE(sibling));  
    file_data_free(vFileData); g_free(path); g_free(name); return;
  }

  /* free the old attached data */
  file_data_free(gtk_ctree_node_get_row_data(GTK_CTREE(pGFileTree->Ctree),sibling));

  /* change text */
  gtk_ctree_get_node_info(GTK_CTREE(pGFileTree->Ctree),sibling,&text,
    &spacing,&pixmap_closed,&mask_closed,&pixmap_opened,&mask_opened,
    &is_leaf,&expanded);

  gtk_ctree_set_node_info(GTK_CTREE(pGFileTree->Ctree),sibling,name,
    spacing,pixmap_closed,mask_closed,pixmap_opened,mask_opened,
    is_leaf,expanded);

  /* attach new data */
  gtk_ctree_node_set_row_data(GTK_CTREE(pGFileTree->Ctree),sibling,vFileData);

  g_free(path); g_free(name);
}
/************************************************************/

/*************************************************************/
/* pFileDataList is the new content of the current directory */
void gfiletree_update(GFileTree *pGFileTree, FileDataList *pFileDataList)
{
  gchar *text;
  guint8 spacing;
  GdkPixmap *pixmap_cb;
  GdkBitmap *mask_cb;
  GtkCTreeNode *try_node,*sibling,*try_node_next;
  GtkCTreeNode *parent = NULL;
  
  FileType *normal_dir_type,*normal_open_dir_type;
  gchar *path;
  guint start,end;
  FileData *vFileData;
  FileDataList *vFileDataList;
  GList *data_node,*data_node_last;
  gboolean under_scan_needed = FALSE;
  GSList *node_list = NULL;
  GSList *node_list_next = NULL;

  file_data_list_init(pFileDataList);
  /* if the list is empty then exit */
  if(file_data_list_is_end(pFileDataList) == TRUE) return;

  if((try_node = gfiletree_find_path(pGFileTree,
      file_data_list_current(pFileDataList)->path)) == NULL) return;
  /*g_print("tree update node found\n");*/
  
  parent = try_node;
  vFileData = gtk_ctree_node_get_row_data(pGFileTree->Ctree,parent);
  if(vFileData->extra_info & FILE_DATA_IS_DIR_SCAN) {
    under_scan_needed = TRUE;
  } else {
    vFileData = gtk_ctree_node_get_row_data(pGFileTree->Ctree,
      GTK_CTREE_ROW(parent)->parent);
    /* if the part we want to update was never scan, forget about it */
    if(!(vFileData->extra_info & FILE_DATA_IS_DIR_SCAN)) return;
  }
  
  /* create a FileDataList with only directory */
  vFileDataList = file_data_list_new();
  file_data_list_init(pFileDataList);
  while(!file_data_list_is_end(pFileDataList))
  {
    vFileData = file_data_list_current(pFileDataList);
    if(gfiletree_filter(pGFileTree, vFileData))
      file_data_list_append(vFileDataList,vFileData);
    file_data_list_next(pFileDataList);
  }

  /* FileType that we might need */
  normal_dir_type = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    NORMAL_DIRECTORY,FALSE);
  if(normal_dir_type == NULL) {gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}
  normal_open_dir_type = file_type_list_find_file_type(pGFileTree->Preference.file_type_list,
    NORMAL_OPEN_DIRECTORY,FALSE);
  if(normal_open_dir_type == NULL){gfile_error(FILE_TYPE_ERROR); exit(FILE_TYPE_ERROR);}

  /* generate tree node list corresponding to our directory */
  try_node = GTK_CTREE_ROW(try_node)->children;
  while(((try_node)&&(GTK_CTREE_ROW(try_node)->parent == parent))||
        gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),parent,try_node))
  {
    while((!((try_node)&&(GTK_CTREE_ROW(try_node)->parent == parent))) &&
          gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),parent,try_node))
    { try_node = GTK_CTREE_NODE_NEXT(try_node); }

    node_list_next = node_list;
    node_list = g_malloc(sizeof(GSList));
    node_list->next = node_list_next;
    node_list->data = try_node;
  
    try_node = GTK_CTREE_NODE_NEXT(try_node);
  }
  
  while(node_list)
  {
    try_node = node_list->data;
    node_list_next = node_list->next;
    g_free(node_list);
    
    vFileData = gtk_ctree_node_get_row_data(pGFileTree->Ctree,try_node);

    /*g_print("<> searching tree %s\n",vFileData->name);*/
	
    data_node_last = vFileDataList->list;
    data_node = vFileDataList->list->next;
    while(data_node &&
          !file_data_is_equal(vFileData,(FileData *)data_node->data))
    { data_node_last = data_node; data_node = data_node->next; }
    
    if(!data_node) {
      /* the item present in the tree no more exist */
      /*g_print("<> %s will be remove\n",vFileData->name);*/
      gtk_ctree_post_recursive(GTK_CTREE(pGFileTree->Ctree), 
        GTK_CTREE_NODE(try_node),GTK_CTREE_FUNC(free_node_data),NULL);
      gtk_ctree_remove_node(GTK_CTREE(pGFileTree->Ctree),GTK_CTREE_NODE(try_node));
    } else {

      /* if stat have changed, del all and rebuild */
      if(!((vFileData->stat.st_dev     == (((FileData *)data_node->data)->stat).st_dev) &&
           (vFileData->stat.st_ino     == (((FileData *)data_node->data)->stat).st_ino) &&
           (vFileData->stat.st_mode    == (((FileData *)data_node->data)->stat).st_mode) &&
           (vFileData->stat.st_nlink   == (((FileData *)data_node->data)->stat).st_nlink) &&
           (vFileData->stat.st_uid     == (((FileData *)data_node->data)->stat).st_uid) &&
           (vFileData->stat.st_gid     == (((FileData *)data_node->data)->stat).st_gid) &&
           (vFileData->stat.st_rdev    == (((FileData *)data_node->data)->stat).st_rdev) &&
           (vFileData->stat.st_size    == (((FileData *)data_node->data)->stat).st_size) &&
           (vFileData->stat.st_blksize == (((FileData *)data_node->data)->stat).st_blksize) &&
           (vFileData->stat.st_blocks  == (((FileData *)data_node->data)->stat).st_blocks) &&
           (vFileData->stat.st_mtime   == (((FileData *)data_node->data)->stat).st_mtime) &&
           (vFileData->stat.st_ctime   == (((FileData *)data_node->data)->stat).st_ctime)))
      {
        /*g_print("<> stat changed for %s\n",vFileData->name);*/
        /* remove */
        gtk_ctree_post_recursive(GTK_CTREE(pGFileTree->Ctree), 
          GTK_CTREE_NODE(try_node),GTK_CTREE_FUNC(free_node_data),NULL);
        gtk_ctree_remove_node(GTK_CTREE(pGFileTree->Ctree),GTK_CTREE_NODE(try_node));
        /* rebuild */
        sibling = gtk_ctree_insert_node(GTK_CTREE(pGFileTree->Ctree),
          parent,NULL,&((FileData *)data_node->data)->name,5,
          normal_dir_type->pixmap,normal_dir_type->mask,
          normal_open_dir_type->pixmap,normal_open_dir_type->mask, FALSE,FALSE);
        gtk_ctree_node_set_row_data(GTK_CTREE(pGFileTree->Ctree),sibling,
          file_data_copy((FileData *)data_node->data));
    
        /* under scan if needed */
        if(under_scan_needed)
        {
          path = g_strconcat(((FileData *)data_node->data)->path,
            ((FileData *)data_node->data)->name,NULL);
          gfiletree_load_path(pGFileTree,path,sibling,0);
          g_free(path);
        }
      }
      /*g_print("<> suppress in vFileDataList %s\n",
        ((FileData *)data_node->data)->name);*/
      /* suppress this item from our vFileDataList */
      data_node_last->next = data_node->next;
      file_data_free((FileData *)data_node->data);
      g_free(data_node);
    }
    node_list = node_list_next;
  }
  
  /* Add all item still present in the vFileDataList because they were
     not in the tree before. */
  file_data_list_init(vFileDataList);
  while(!file_data_list_is_end(vFileDataList))
  {
    /*g_print("<> new directory %s\n",file_data_list_current(vFileDataList)->name);*/
    /* add the new item */
    sibling = gtk_ctree_insert_node(GTK_CTREE(pGFileTree->Ctree),
      parent,NULL,&file_data_list_current(vFileDataList)->name,5,
      normal_dir_type->pixmap,
      normal_dir_type->mask,
      normal_open_dir_type->pixmap,
      normal_open_dir_type->mask, FALSE,FALSE);
    gtk_ctree_node_set_row_data(GTK_CTREE(pGFileTree->Ctree),sibling,
      file_data_copy(file_data_list_current(vFileDataList)));
    
    /* under scan if needed */
    if(under_scan_needed)
    {
      path = g_strconcat(file_data_list_current(vFileDataList)->path,
        file_data_list_current(vFileDataList)->name,NULL);
      gfiletree_load_path(pGFileTree,path,sibling,0);
      g_free(path);
    }
			     
    file_data_list_next(vFileDataList);
  }
     
  /* free the allocated FileDataList */
  file_data_list_free(vFileDataList);
}
/*************************************************************/

/************************************************************************/
/* Return the tree node corresponding to path. Return NULL if not found */
/* path can be ended by a '/' or nothing.                               */
static GtkCTreeNode *gfiletree_find_path(GFileTree *pGFileTree, gchar *path)
{
  gchar *text;
  guint8 spacing;
  GdkPixmap *pixmap_cb;
  GdkBitmap *mask_cb;
  GtkCTreeNode *try_node;
  GtkCTreeNode *parent = NULL;
  guint start,end;
  
  start = 1;
  try_node = pGFileTree->root_node;

  /* try to find the node corresponding to the path */  
  while(path[start])
  {
    for(end=start;(path[end] != '/') && (path[end]); end++);

    parent = try_node;
    try_node = GTK_CTREE_ROW(try_node)->children;

    while(((try_node)&&(GTK_CTREE_ROW(try_node)->parent==parent))||
          gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),parent,try_node))
    {

      while((!((try_node)&&(GTK_CTREE_ROW(try_node)->parent == parent))) &&
            gtk_ctree_is_ancestor(GTK_CTREE(pGFileTree->Ctree),parent,try_node))
      { try_node = GTK_CTREE_NODE_NEXT(try_node); }

      gtk_ctree_node_get_pixtext(pGFileTree->Ctree,try_node,0,&text,
        &spacing,&pixmap_cb,&mask_cb);
      /* if equal then we have found the node */
      if((strlen(text) == end-start) && 
         (!strncmp(&path[start],text,end-start))) break;
 
      try_node = GTK_CTREE_NODE_NEXT(try_node);
    }
    /* if not found, exit */
    if((!try_node)||(GTK_CTREE_ROW(try_node)->parent != parent)) return NULL;
    /* continue with next directory */
    start = end;
    /* if we stoped on a '/' jump over it */
    if(path[end]) start++;
  }
  /* if we are here, we have found the node corresponding to the path */
  return try_node;
}
/************************************************************************/

/*******************************************************/
/* create a list of all the selected file in the ctree */
/* when we are called = the current_Path.	       */
/* return : The list.                                  */
FileDataList *gfiletree_list_selected(GFileTree *pGFileTree)
{ FileData *vFileData;
  FileDataList *vFileDataList;
  
  vFileDataList = file_data_list_new();
  /* get the data attached to the current GtkCtreeNode */
  vFileData = gtk_ctree_node_get_row_data(GTK_CTREE(pGFileTree->Ctree),pGFileTree->last_node);
  /* append the data to the list */
  file_data_list_append(vFileDataList, vFileData);
  /* return the list */    
  return(vFileDataList);
}
/*******************************************************/

/****************************************/
/* for the popup menu                   */
/* ________ UNDER DEVELOPEMENT _________*/  
static gint gfiletree_event_cb(GtkCList *clist, GdkEvent *event,
  GFileTree *pGFileTree)
{ gint row,column;
  FileDataList *vFileDataList;

  if(event->type == GDK_BUTTON_PRESS) {
    GdkEventButton *bevent = (GdkEventButton *) event; 
    /* if it is the third button */
    if(bevent->button == 3)
    { /*g_print("properties called.\n");*/
      if(pGFileTree->Preference.pop_up_prop_menu == FALSE) return(FALSE);

      row = bevent->x; column = bevent->y;		      
      gtk_clist_get_selection_info(GTK_CLIST(pGFileTree->Ctree),
        row,column,&row,&column);
      /*g_print("row : %d, col : %d\n",row,column);*/

      /* do something only if the user right click on a selected file */
/*      if(gfilelist_list_is_row_selected(pGFileList,row) == TRUE)*/
/*      if(gtk_ctree_is_hot_spot(GTK_CTREE(pGFileTree->Ctree),bevent->x,bevent->y) == TRUE)
*/      if(gtk_clist_get_row_data(GTK_CLIST(pGFileTree->Ctree),row) ==
         gtk_ctree_node_get_row_data(GTK_CTREE(pGFileTree->Ctree),pGFileTree->last_node))
      {
        vFileDataList = gfiletree_list_selected(pGFileTree);
        pop_menu_new(vFileDataList,pGFileTree->current_path,bevent->button, bevent->time);
	file_data_list_free(vFileDataList);
      } else {
        pop_menu_new(NULL,pGFileTree->current_path,bevent->button, bevent->time);
      }
      /* Tell calling code that we have handled this event; the buck
      * stops here. */	      
      return TRUE;      
    }  	      
    /* Tell calling code that we have not handled this event; pass it on. */
    return FALSE;
  }
  /* Tell calling code that we have not handled this event; pass it on. */
  return FALSE;
}
/****************************************/

/*******************************************/
/* UNDER DEVELOPPEMENT FOR THE DRAG & DROP */
/* not done                                */
/*******************************************/

static void source_drag_data_get(GtkWidget *widget,GdkDragContext *context,
  GtkSelectionData *selection_data,guint info,guint time,GFileTree *pGFileTree)
{ 
  FileDataList *vFileDataList;
  gchar *string,*tmp_string,*tmp_string2;

  /*g_print("get_data called.\n");*/
  if(info == TARGET_FILES) {
    vFileDataList = gfiletree_list_selected(pGFileTree);
    file_data_list_init(vFileDataList);
    if(file_data_list_is_end(vFileDataList) == FALSE)
    {
      string = g_strdup("");
      while(file_data_list_is_end(vFileDataList) == FALSE)
      {
        tmp_string = g_strdup_printf("file:%s%s\n",
          file_data_list_current(vFileDataList)->path,
          file_data_list_current(vFileDataList)->name);
	tmp_string2 = string;
	string = g_strconcat(string,tmp_string,NULL);
	g_free(tmp_string); g_free(tmp_string2);
	file_data_list_next(vFileDataList);
      }
      gtk_selection_data_set(selection_data,selection_data->target,
        8, string, strlen(string));
      g_free(string);
    }
  } else {
    g_print("source data get NOT TARGET_FILES\n");
  }
}
